# test on compton module
import pytest
import numpy as np
import astropy.units as u
from astropy.constants import M_sun
from astropy.coordinates import Distance
from pathlib import Path
from agnpy.emission_regions import Blob
from agnpy.targets import (
    PointSourceBehindJet,
    SSDisk,
    SphericalShellBLR,
    RingDustTorus,
    CMB,
)
from agnpy.compton import SynchrotronSelfCompton, ExternalCompton
from .utils import make_comparison_plot, extract_columns_sample_file, check_deviation
from agnpy.utils.math import trapz_loglog


agnpy_dir = Path(__file__).parent.parent
# where to read sampled files
data_dir = agnpy_dir / "data"
# where to save figures
figures_dir = agnpy_dir.parent / "crosschecks/figures/compton"
Path(figures_dir / "ssc").mkdir(parents=True, exist_ok=True)
for subdir in ["disk", "blr", "dt", "cmb"]:
    Path(figures_dir / "ec" / subdir).mkdir(parents=True, exist_ok=True)

# variables with _test are global and meant to be used in all tests
pwl_spectrum_norm_test = 1e48 * u.Unit("erg")
pwl_dict_test = {
    "type": "PowerLaw",
    "parameters": {"p": 2.8, "gamma_min": 1e2, "gamma_max": 1e5},
}
bpwl_spectrum_norm_test = 6e42 * u.Unit("erg")
bpwl_dict_test = {
    "type": "BrokenPowerLaw",
    "parameters": {
        "p1": 2.0,
        "p2": 3.5,
        "gamma_b": 1e4,
        "gamma_min": 20,
        "gamma_max": 5e7,
    },
}
# blob reproducing Figure 7.4 of Dermer Menon 2009
pwl_blob_test = Blob(
    1e16 * u.cm,
    Distance(1e27, unit=u.cm).z,
    10,
    10,
    1 * u.G,
    pwl_spectrum_norm_test,
    pwl_dict_test,
)
# blob reproducing the EC scenarios in Finke 2016
bpwl_blob_test = Blob(
    1e16 * u.cm, 1, 40, 40, 0.56 * u.G, bpwl_spectrum_norm_test, bpwl_dict_test,
)
bpwl_blob_test.set_gamma_size(400)
# global disk
M_BH = 1.2 * 1e9 * M_sun.cgs
L_disk = 2e46 * u.Unit("erg s-1")
eta = 1 / 12
R_in = 6
R_out = 200
disk_test = SSDisk(M_BH, L_disk, eta, R_in, R_out, R_g_units=True)
# global blr
xi_line = 0.024
R_line = 1e17 * u.cm
blr_test = SphericalShellBLR(L_disk, xi_line, "Lyalpha", R_line)
# global dt
T_dt = 1e3 * u.K
csi_dt = 0.1
dt_test = RingDustTorus(L_disk, csi_dt, T_dt)


class TestSynchrotronSelfCompton:
    """class grouping all tests related to the Synchrotron Slef Compton class"""

    @pytest.mark.parametrize("gamma_max, nu_range_max", [("1e5", 1e25), ("1e7", 1e27)])
    def test_ssc_reference_sed(self, gamma_max, nu_range_max):
        """test agnpy SSC SED against the ones in Figure 7.4 of Dermer Menon 2009"""
        # reference SED
        nu_ref, sed_ref = extract_columns_sample_file(
            f"{data_dir}/reference_seds/dermer_menon_2009/figure_7_4/ssc_gamma_max_{gamma_max}.txt",
            "Hz",
            "erg cm-2 s-1",
        )
        # agnpy
        # change the gamma_max in the blob
        pwl_dict_test["parameters"]["gamma_max"] = float(gamma_max)
        pwl_blob_test.set_spectrum(pwl_spectrum_norm_test, pwl_dict_test, "integral")
        # recompute the SED at the same ordinates where the figure was sampled
        ssc = SynchrotronSelfCompton(pwl_blob_test)
        sed_agnpy = ssc.sed_flux(nu_ref)
        # sed comparison plot
        nu_range = [1e14, nu_range_max] * u.Hz
        make_comparison_plot(
            nu_ref,
            sed_agnpy,
            sed_ref,
            "agnpy",
            "Figure 7.4, Dermer and Menon (2009)",
            "Synchrotron Self Compton, " + r"$\gamma_{max} = $" + gamma_max,
            f"{figures_dir}/ssc/comparison_gamma_max_{gamma_max}_figure_7_4_dermer_menon_2009.png",
            "sed",
            y_range=[1e-13, 1e-9],
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 20% from the figure
        assert check_deviation(nu_ref, sed_agnpy, sed_ref, 0.2, nu_range)

    def test_ssc_integration_methods(self):
        """test SSC SED for different integration methods against each other
        """
        nu = np.logspace(15, 28) * u.Hz
        ssc_trapz = SynchrotronSelfCompton(pwl_blob_test, integrator=np.trapz)
        ssc_trapz_loglog = SynchrotronSelfCompton(
            pwl_blob_test, integrator=trapz_loglog
        )
        sed_ssc_trapz = ssc_trapz.sed_flux(nu)
        sed_ssc_trapz_loglog = ssc_trapz_loglog.sed_flux(nu)
        # compare in a restricted energy range
        nu_range = [1e15, 1e27] * u.Hz
        make_comparison_plot(
            nu,
            sed_ssc_trapz_loglog,
            sed_ssc_trapz,
            "trapezoidal log-log integration",
            "trapezoidal integration",
            "Synchrotron Self Compton",
            f"{figures_dir}/ssc/comparison_integration_methods.png",
            "sed",
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 15%
        assert check_deviation(nu, sed_ssc_trapz_loglog, sed_ssc_trapz, 0.15, nu_range)


class TestExternalCompton:
    """class grouping all tests related to the External Compton class"""

    # tests for EC on SSDisk
    @pytest.mark.parametrize("r", ["1e17", "1e18"])
    def test_ec_disk_reference_sed(self, r):
        """test agnpy SED for EC on Disk against the one in Figure 8 of Finke 2016"""
        # reference SED
        nu_ref, sed_ref = extract_columns_sample_file(
            f"{data_dir}/reference_seds/finke_2016/figure_8/ec_disk_r_{r}.txt",
            "Hz",
            "erg cm-2 s-1",
        )
        # recompute the SED at the same ordinates where the figure was sampled
        ec_disk = ExternalCompton(bpwl_blob_test, disk_test, float(r) * u.cm)
        sed_agnpy = ec_disk.sed_flux(nu_ref)
        # check in a restricted energy range
        nu_range = [1e18, 1e28] * u.Hz
        make_comparison_plot(
            nu_ref,
            sed_agnpy,
            sed_ref,
            "agnpy",
            "Figure 8, Finke (2016)",
            f"External Compton on Shakura Sunyaev Disk, r = {r} cm",
            f"{figures_dir}/ec/disk/comparison_r_{r}_cm_figure_8_finke_2016.png",
            "sed",
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 35% from the figure
        assert check_deviation(nu_ref, sed_agnpy, sed_ref, 0.35, nu_range)

    def test_ec_disk_integration_methods(self):
        """test EC on Disk SED for different integration methods against each other
        """
        nu = np.logspace(15, 28) * u.Hz
        r = 1e18 * u.cm
        ec_disk_trapz = ExternalCompton(
            bpwl_blob_test, disk_test, r, integrator=np.trapz
        )
        ec_disk_trapz_loglog = ExternalCompton(
            bpwl_blob_test, disk_test, r, integrator=trapz_loglog
        )
        sed_ec_disk_trapz = ec_disk_trapz.sed_flux(nu)
        sed_ec_disk_trapz_loglog = ec_disk_trapz_loglog.sed_flux(nu)
        make_comparison_plot(
            nu,
            sed_ec_disk_trapz_loglog,
            sed_ec_disk_trapz,
            "trapezoidal log-log integration",
            "trapezoidal integration",
            "External Compton on Shakura Sunyaev Disk",
            f"{figures_dir}/ec/disk/comparison_integration_methods.png",
            "sed",
        )
        # requires that the SED points deviate less than 20%
        assert check_deviation(nu, sed_ec_disk_trapz_loglog, sed_ec_disk_trapz, 0.2)

    # tests for EC on BLR
    @pytest.mark.parametrize("r", ["1e18", "1e19"])
    def test_ec_blr_reference_sed(self, r):
        """test agnpy SED for EC on BLR against the one in Figure 10 of Finke 2016"""
        # reference SED
        nu_ref, sed_ref = extract_columns_sample_file(
            f"{data_dir}/reference_seds/finke_2016/figure_10/ec_blr_r_{r}.txt",
            "Hz",
            "erg cm-2 s-1",
        )
        # recompute the SED at the same ordinates where the figure was sampled
        ec_blr = ExternalCompton(bpwl_blob_test, blr_test, float(r) * u.cm)
        sed_agnpy = ec_blr.sed_flux(nu_ref)
        # check in a restricted energy range
        nu_range = [1e18, 1e28] * u.Hz
        make_comparison_plot(
            nu_ref,
            sed_agnpy,
            sed_ref,
            "agnpy",
            "Figure 10, Finke (2016)",
            f"External Compton on Spherical Shell BLR, r = {r} cm",
            f"{figures_dir}/ec/blr/comparison_r_{r}_cm_figure_10_finke_2016.png",
            "sed",
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 30% from the figure
        assert check_deviation(nu_ref, sed_agnpy, sed_ref, 0.3, nu_range)

    def test_ec_blr_integration_methods(self):
        """test EC on BLR SED for different integration methods
        """
        nu = np.logspace(15, 28) * u.Hz
        r = 1e18 * u.cm
        ec_blr_trapz = ExternalCompton(bpwl_blob_test, blr_test, r, integrator=np.trapz)
        ec_blr_trapz_loglog = ExternalCompton(
            bpwl_blob_test, blr_test, r, integrator=trapz_loglog
        )
        sed_ec_blr_trapz = ec_blr_trapz.sed_flux(nu)
        sed_ec_blr_trapz_loglog = ec_blr_trapz_loglog.sed_flux(nu)
        # check in a restricted energy range
        make_comparison_plot(
            nu,
            sed_ec_blr_trapz_loglog,
            sed_ec_blr_trapz,
            "trapezoidal log-log integration",
            "trapezoidal integration",
            "External Compton on Spherical Shell Broad Line Region",
            f"{figures_dir}/ec/blr/comparison_integration_methods.png",
            "sed",
        )
        # requires that the SED points deviate less than 30%
        assert check_deviation(nu, sed_ec_blr_trapz_loglog, sed_ec_blr_trapz, 0.3)

    # tests for EC on DT
    @pytest.mark.parametrize("r", ["1e20", "1e21"])
    def test_ec_dt_reference_sed(self, r):
        """test agnpy SED for EC on DT against the one in Figure 11 of Finke 2016"""
        # reference SED
        nu_ref, sed_ref = extract_columns_sample_file(
            f"{data_dir}/reference_seds/finke_2016/figure_11/ec_dt_r_{r}.txt",
            "Hz",
            "erg cm-2 s-1",
        )
        # recompute the SED at the same ordinates where the figure was sampled
        ec_dt = ExternalCompton(bpwl_blob_test, dt_test, float(r) * u.cm)
        sed_agnpy = ec_dt.sed_flux(nu_ref)
        # check in a restricted energy range
        nu_range = [1e18, 1e28] * u.Hz
        make_comparison_plot(
            nu_ref,
            sed_agnpy,
            sed_ref,
            "agnpy",
            "Figure 11, Finke (2016)",
            f"External Compton on Ring Dust Torus, r = {r} cm",
            f"{figures_dir}/ec/dt/comparison_r_{r}_cm_figure_11_finke_2016.png",
            "sed",
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 30% from the figure
        assert check_deviation(nu_ref, sed_agnpy, sed_ref, 0.3, nu_range)

    def test_ec_dt_integration_methods(self):
        """test EC on DT SED for different integration methods
        """
        nu = np.logspace(15, 28) * u.Hz
        r = 1e20 * u.cm
        ec_dt_trapz = ExternalCompton(bpwl_blob_test, dt_test, r, integrator=np.trapz)
        ec_dt_trapz_loglog = ExternalCompton(
            bpwl_blob_test, dt_test, r, integrator=trapz_loglog
        )
        sed_ec_dt_trapz = ec_dt_trapz.sed_flux(nu)
        sed_ec_dt_trapz_loglog = ec_dt_trapz_loglog.sed_flux(nu)
        make_comparison_plot(
            nu,
            sed_ec_dt_trapz_loglog,
            sed_ec_dt_trapz,
            "trapezoidal log-log integration",
            "trapezoidal integration",
            "External Compton on Ring Dust Torus",
            f"{figures_dir}/ec/dt/comparison_integration_methods.png",
            "sed",
        )
        # requires that the SED points deviate less than 20%
        assert check_deviation(nu, sed_ec_dt_trapz_loglog, sed_ec_dt_trapz, 0.2)

    # tests against point-like sources approximating the targets
    def test_ec_blr_vs_point_source(self):
        """check if in the limit of large distances the EC on the BLR tends to
        the one of a point-like source approximating it"""
        r = 1e22 * u.cm
        # point like source approximating the blr
        ps_blr = PointSourceBehindJet(
            blr_test.xi_line * blr_test.L_disk, blr_test.epsilon_line
        )
        # external Compton
        ec_blr = ExternalCompton(bpwl_blob_test, blr_test, r)
        ec_ps_blr = ExternalCompton(bpwl_blob_test, ps_blr, r)
        # seds
        nu = np.logspace(15, 30) * u.Hz
        sed_ec_blr = ec_blr.sed_flux(nu)
        sed_ec_ps_blr = ec_ps_blr.sed_flux(nu)
        # sed comparison plot
        make_comparison_plot(
            nu,
            sed_ec_ps_blr,
            sed_ec_blr,
            "point source approximating the BLR",
            "spherical shell BLR",
            "External Compton on Spherical Shell BLR, "
            + r"$r = 10^{22}\,{\rm cm} \gg R_{\rm line}$",
            f"{figures_dir}/ec/blr/comparison_point_source.png",
            "sed",
        )
        # requires a 20% deviation from the two SED points
        assert check_deviation(nu, sed_ec_ps_blr, sed_ec_blr, 0.2)

    def test_ec_dt_vs_point_source(self):
        """check if in the limit of large distances the EC on the DT tends to
        the one of a point-like source approximating it"""
        r = 1e22 * u.cm
        # point like source approximating the dt
        ps_dt = PointSourceBehindJet(dt_test.xi_dt * dt_test.L_disk, dt_test.epsilon_dt)
        # external Compton
        ec_dt = ExternalCompton(bpwl_blob_test, dt_test, r)
        ec_ps_dt = ExternalCompton(bpwl_blob_test, ps_dt, r)
        # seds
        nu = np.logspace(15, 28) * u.Hz
        sed_ec_dt = ec_dt.sed_flux(nu)
        sed_ec_ps_dt = ec_ps_dt.sed_flux(nu)
        make_comparison_plot(
            nu,
            sed_ec_ps_dt,
            sed_ec_dt,
            "point source approximating the DT",
            "ring Dust Torus",
            "External Compton on Ring Dust Torus, "
            + r"$r = 10^{22}\,{\rm cm} \gg R_{\rm dt}$",
            f"{figures_dir}/ec/dt/comparison_point_source.png",
            "sed",
        )
        # requires a 20% deviation from the two SED points
        assert check_deviation(nu, sed_ec_dt, sed_ec_ps_dt, 0.2)

    def test_ec_cmb_vs_jetset(self):
        """check the SED for EC on CMB against jetset"""
        # reference SED
        file_ref = f"{data_dir}/reference_seds/jetset/data/ec_cmb_bpwl_jetset_1.1.2.txt"
        nu_ref, sed_ref = extract_columns_sample_file(file_ref, "Hz", "erg cm-2 s-1")
        # recompute the SED at the same ordinates where the figure was sampled
        cmb = CMB(z=bpwl_blob_test.z)
        ec_cmb = ExternalCompton(bpwl_blob_test, cmb)
        sed_agnpy = ec_cmb.sed_flux(nu_ref)
        # sed comparison plot, we will check between 10^(11) and 10^(19) Hz
        nu_range = [1e16, 5e27] * u.Hz
        make_comparison_plot(
            nu_ref,
            sed_agnpy,
            sed_ref,
            "agnpy",
            "jetset 1.1.2",
            "EC on CMB, z = 1",
            f"{figures_dir}/ec/cmb/ec_cmb_bpwl_comparison_jetset_1.1.2.png",
            "sed",
            comparison_range=nu_range.to_value("Hz"),
        )
        # requires that the SED points deviate less than 35% from the figure
        assert check_deviation(nu_ref, sed_agnpy, sed_ref, 0.35, nu_range)



import numpy as np
import astropy.units as u
from astropy.constants import e, sigma_T
from ..utils.conversion import mec2, B_to_cgs

e = e.gauss

__all__ = ["SpectralConstraints"]


class SpectralConstraints:
    r""" Class to describe the self-consistency constraints on the electron energy distribution

    Parameters
    ----------
    blob : :class: `Blob` emission region
    """

    def __init__(self, blob):
        self.blob = blob

    @property
    def gamma_max_larmor(self):
        r"""maximum Lorentz factor of electrons that have their Larmour radius 
        smaller than the blob radius: :math:`R_L < R_b`. 
        The Larmor frequency and radius in Gaussian units read

        .. math::

            \omega_L &= \frac{eB}{\gamma m_e c} \\
            R_L &= \frac{v}{\omega_L} = \frac{\gamma m_e v c}{e B} \approx \frac{\gamma m_e c^2}{e B}

        therefore

        .. math::

            R_L < R_b \Rightarrow \gamma_{\mathrm{max}} < \frac{R_b e B}{m_e c^2}
        """
        return (self.blob.R_b * e * self.blob.B_cgs / mec2).to_value("")

    @property
    def gamma_max_ballistic(self):
        r"""Naive estimation of maximum Lorentz factor of electrons comparing 
        acceleration time scale with ballistic time scale. 
        For the latter we assume that the particles crosses the blob radius.

        .. math::

            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{acc}} &= \xi c E / R_L \\
            T_{\mathrm{acc}} &= E \,/\,(\mathrm{d}E/\mathrm{d}t)_{\mathrm{acc}} = R_L / (\xi c) \\
            T_{\mathrm{bal}} &= R_b / c \\
            T_{\mathrm{acc}} &< T_{\mathrm{bal}} 
            \Rightarrow \gamma_{\mathrm{max}} < \frac{\xi  R_b e B}{m_e c^2} 
        """
        return self.blob.xi * self.gamma_max_larmor

    @property
    def gamma_max_synch(self):
        r"""Simple estimation of maximum Lorentz factor of electrons 
        comparing the acceleration time scale with the synchrotron energy loss

        .. math::
            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{acc}} &= \xi c E / R_L \\
            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{synch}} &= 4 / 3 \sigma_T c U_B \gamma^2 \\
            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{acc}} &= (\mathrm{d}E/\mathrm{d}t)_{\mathrm{synch}} 
            \Rightarrow \gamma_{\mathrm{max}} < \sqrt{\frac{6 \pi \xi e}{\sigma_T B}}
        """
        return np.sqrt(
            6 * np.pi * self.blob.xi * e / (sigma_T * self.blob.B_cgs)
        ).to_value("")

    @property
    def gamma_max_SSC(self):
        r"""Simple estimation of maximum Lorentz factor of electrons 
        comparing the acceleration time scale with the SSC energy loss (in Thomson range)
        WARNING: the highest energy electrons will most often scatter in Klein-Nishina range instead

        .. math::
            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{acc}} &= \xi c E / R_L \\
            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{SSC}} &= 4 / 3 \sigma_T c U_{\mathrm{synch}} \gamma^2 \\
            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{acc}} &= (\mathrm{d}E/\mathrm{d}t)_{\mathrm{SSC}} 
            \Rightarrow \gamma_{\mathrm{max}} < \sqrt{\frac{3 \xi e B }{\sigma_T U_SSC}}
        """
        return np.sqrt(
            3
            * self.blob.xi
            * e
            * self.blob.B_cgs
            / (4 * sigma_T * self.blob.u_ph_synch)
        ).to_value("")

    def gamma_max_EC_DT(self, dt, r=0 * u.cm):
        r"""Simple estimation of maximum Lorentz factor of electrons comparing the acceleration time scale 
        with the EC energy loss (in Thomson range, see B&G 1970), like in gamma_max_SSC 
        WARNING: assumes Thomson regime

        .. math::            
            \gamma_{\mathrm{max}} = \sqrt{\frac{3 \xi e B }{ \sigma_T U'_\mathrm{ext}}} 
        """
        return np.sqrt(
            3 * self.blob.xi * e * self.blob.B_cgs / (4 * sigma_T * dt.u(r, self.blob))
        ).to_value("")

    @property
    def gamma_break_synch(self):
        r"""Simple estimation of the cooling break of electrons comparing 
        synchrotron cooling time scale with the ballistic time scale: 
        
        .. math::

            T_{\mathrm{synch}} &= E\,/\,(\mathrm{d}E/\mathrm{d}t)_{\mathrm{synch}} 
            =  3 m_e c^2 / (4 \sigma_T U_B \gamma) \\
            T_{\mathrm{bal}} &= R_b / c \\
            T_{\mathrm{synch}} &= T_{\mathrm{bal}} \Rightarrow \gamma_b = 6 \pi m_e c^2 / \sigma_T B^2 R_b 
        """
        gamma_max = (
            (
                6
                * np.pi
                * mec2
                / (sigma_T * np.power(self.blob.B_cgs, 2) * self.blob.R_b)
            )
            .to("")
            .value
        )
        return gamma_max

    @property
    def gamma_break_SSC(self):
        r"""Simple estimation of the cooling break of electrons comparing 
        SSC time scale (see B&G 1970) with the ballistic time scale: 
        WARNING: only applicable in Thomson regime

        .. math::
            T_{\mathrm{SSC}} &= E\,/\,(\mathrm{d}E/\mathrm{d}t)_{\mathrm{SSC}} 
            =  3 m_e c^2 / (4 \sigma_T U_{\mathrm{SSC}} \gamma) \\
            T_{\mathrm{bal}} &= R_b / c \\
            T_{\mathrm{SSC}} &= T_{\mathrm{bal}} \Rightarrow \gamma_b = 3  m_e c^2 / 4 \sigma_T U_{\mathrm{SSC}} R_b 
        """
        return (
            (3 * mec2 / (4 * sigma_T * self.blob.u_ph_synch * self.blob.R_b))
            .to("")
            .value
        )

    def gamma_break_EC_DT(self, dt, r=0 * u.cm):
        r"""Simple estimation of the cooling break of electrons comparing 
        EC time scale (see B&G 1970) with the ballistic time scale, like in gamma_break_SSC 
        WARNING: assumes Thomson regime

        .. math::            
            \gamma_b = 3  m_e c^2 / 4 \sigma_T U'_{\mathrm{ext}} R_b 
        """
        #        u_ext=np.power(self.Gamma,2) * np.power(1-mu*self.Beta,2) * dt.xi_dt*dt.L_disk/(4*np.pi*np.power(d,2) * c)
        return (
            (3 * mec2 / (4 * sigma_T * dt.u(r, self.blob) * self.blob.R_b)).to("").value
        )



# module containing the targets for the external compton radiation
import numpy as np
from astropy.constants import c, G, M_sun, k_B, sigma_sb
import astropy.units as u
from astropy.coordinates import Distance
from astropy.modeling.models import BlackBody
from ..utils.conversion import mec2, nu_to_epsilon_prime
from ..utils.math import axes_reshaper


__all__ = [
    "lines_dictionary",
    "CMB",
    "SSDisk",
    "PointSourceBehindJet",
    "SphericalShellBLR",
    "RingDustTorus",
]


# dictionary with all the available spectral lines
lines_dictionary = {
    "Lyepsilon": {
        "lambda": 937.80 * u.Angstrom,
        "R_Hbeta_ratio": 2.7,
        "L_Hbeta_ratio": 0.24,
    },
    "Lydelta": {
        "lambda": 949.74 * u.Angstrom,
        "R_Hbeta_ratio": 2.8,
        "L_Hbeta_ratio": 0.24,
    },
    "CIII": {
        "lambda": 977.02 * u.Angstrom,
        "R_Hbeta_ratio": 0.83,
        "L_Hbeta_ratio": 0.60,
    },
    "NIII": {
        "lambda": 990.69 * u.Angstrom,
        "R_Hbeta_ratio": 0.85,
        "L_Hbeta_ratio": 0.60,
    },
    "Lybeta": {
        "lambda": 1025.72 * u.Angstrom,
        "R_Hbeta_ratio": 1.2,
        "L_Hbeta_ratio": 1.1,
    },
    "OVI": {
        "lambda": 1033.83 * u.Angstrom,
        "R_Hbeta_ratio": 1.2,
        "L_Hbeta_ratio": 1.1,
    },
    "ArI": {
        "lambda": 1066.66 * u.Angstrom,
        "R_Hbeta_ratio": 4.5,
        "L_Hbeta_ratio": 0.094,
    },
    "Lyalpha": {
        "lambda": 1215.67 * u.Angstrom,
        "R_Hbeta_ratio": 0.27,
        "L_Hbeta_ratio": 12,
    },
    "OI": {
        "lambda": 1304.35 * u.Angstrom,
        "R_Hbeta_ratio": 4.0,
        "L_Hbeta_ratio": 0.23,
    },
    "SiII": {
        "lambda": 1306.82 * u.Angstrom,
        "R_Hbeta_ratio": 4.0,
        "L_Hbeta_ratio": 0.23,
    },
    "SiIV": {
        "lambda": 1396.76 * u.Angstrom,
        "R_Hbeta_ratio": 0.83,
        "L_Hbeta_ratio": 1.0,
    },
    "OIV]": {
        "lambda": 1402.06 * u.Angstrom,
        "R_Hbeta_ratio": 0.83,
        "L_Hbeta_ratio": 1.0,
    },
    "CIV": {
        "lambda": 1549.06 * u.Angstrom,
        "R_Hbeta_ratio": 0.83,
        "L_Hbeta_ratio": 2.9,
    },
    "NIV": {
        "lambda": 1718.55 * u.Angstrom,
        "R_Hbeta_ratio": 3.8,
        "L_Hbeta_ratio": 0.30,
    },
    "AlII": {
        "lambda": 1721.89 * u.Angstrom,
        "R_Hbeta_ratio": 3.8,
        "L_Hbeta_ratio": 0.30,
    },
    "CIII]": {
        "lambda": 1908.73 * u.Angstrom,
        "R_Hbeta_ratio": 0.46,
        "L_Hbeta_ratio": 1.8,
    },
    "[NeIV]": {
        "lambda": 2423.83 * u.Angstrom,
        "R_Hbeta_ratio": 5.8,
        "L_Hbeta_ratio": 0.051,
    },
    "MgII": {
        "lambda": 2798.75 * u.Angstrom,
        "R_Hbeta_ratio": 0.45,
        "L_Hbeta_ratio": 1.7,
    },
    "Hdelta": {
        "lambda": 4102.89 * u.Angstrom,
        "R_Hbeta_ratio": 3.4,
        "L_Hbeta_ratio": 0.12,
    },
    "Hgamma": {
        "lambda": 4341.68 * u.Angstrom,
        "R_Hbeta_ratio": 3.2,
        "L_Hbeta_ratio": 0.30,
    },
    "HeII": {
        "lambda": 4687.02 * u.Angstrom,
        "R_Hbeta_ratio": 0.63,
        "L_Hbeta_ratio": 0.016,
    },
    "Hbeta": {
        "lambda": 4862.68 * u.Angstrom,
        "R_Hbeta_ratio": 1.0,
        "L_Hbeta_ratio": 1.0,
    },
    "[ClIII]": {
        "lambda": 5539.43 * u.Angstrom,
        "R_Hbeta_ratio": 4.8,
        "L_Hbeta_ratio": 0.039,
    },
    "HeI": {
        "lambda": 5877.29 * u.Angstrom,
        "R_Hbeta_ratio": 0.39,
        "L_Hbeta_ratio": 0.092,
    },
    "Halpha": {
        "lambda": 6564.61 * u.Angstrom,
        "R_Hbeta_ratio": 1.3,
        "L_Hbeta_ratio": 3.6,
    },
}


class CMB:
    """Cosmic Microwave Background radiation, approximated as an isotropic
    monochromatic target.
    
    Parameters
    ----------
    z : float
        redshift at which the CMB is considered
    """

    def __init__(self, z):
        self.name = "Cosmic Microwave Background Radiation"
        a = 7.5657 * 1e-15 * u.Unit("erg cm-3 K-4")  # radiation constant
        T = 2.72548 * u.K
        self.u_0 = (a * np.power(T, 4)).to("erg cm-3") * np.power(1 + z, 4)
        self.epsilon_0 = (2.7 * k_B * T / mec2).to_value("") * (1 + z)

    def u(self, blob=None):
        """integral energy density of the CMB

        Parameters
        ----------
        blob : :class:`~agnpy.emission_regions.Blob`
            if provided, the energy density is computed in a reference frame 
            comvoing with the blob
        """
        if blob:
            return self.u_0 * np.power(blob.Gamma, 2) * (1 + np.power(blob.Beta, 2) / 3)
        else:
            return self.u_0


class PointSourceBehindJet:
    """Monochromatic point source behind the jet.
    
    Parameters
    ----------
    L_0 : :class:`~astropy.units.Quantity`
        luminosity of the source
    epsilon_0 : float
        dimensionless monochromatic energy of the source
    """

    def __init__(self, L_0, epsilon_0):
        self.name = "Monochromatic Point Source Behind the Jet"
        self.L_0 = L_0
        self.epsilon_0 = epsilon_0

    def u(self, r, blob=None):
        """integral energy density of the point source at distance r along the 
        jet axis

        Parameters
        ----------
        r : :class:`~astropy.units.Quantity`
            array of distances along the jet axis
        blob : :class:`~agnpy.emission_regions.Blob`
            if provided, the energy density is computed in a reference frame 
            comvoing with the blob
        """
        u_0 = (self.L_0 / (4 * np.pi * c * np.power(r, 2))).to("erg cm-3")
        if blob:
            return u_0 / (np.power(blob.Gamma, 2) * np.power(1 + blob.Beta, 2))
        else:
            return u_0


class SSDisk:
    """[Shakura1973]_ accretion disk as modeled by [Dermer2002]_

    Parameters
    ----------
    M_BH : :class:`~astropy.units.Quantity`
        Black Hole mass    
    L_disk : :class:`~astropy.units.Quantity`
        luminosity of the disk 
    eta : float
        accretion efficiency
    R_in : :class:`~astropy.units.Quantity` / float
        inner disk radius
    R_out : :class:`~astropy.units.Quantity` / float
        outer disk radius
    R_g_units : bool
        whether or not input radiuses are specified in units of the gravitational radius
    """

    def __init__(self, M_BH, L_disk, eta, R_in, R_out, R_g_units=False):
        self.name = "Shakura Sunyaev Accretion Disk"
        # masses and luminosities
        self.M_BH = M_BH
        self.M_8 = (M_BH / (1e8 * M_sun)).to_value("")
        self.L_Edd = 1.26 * 1e46 * self.M_8 * u.Unit("erg s-1")
        self.L_disk = L_disk
        # fraction of the Eddington luminosity at which the disk is accreting
        self.l_Edd = (self.L_disk / self.L_Edd).to_value("")
        self.eta = eta
        self.m_dot = (self.L_disk / (self.eta * np.power(c, 2))).to("g s-1")
        # gravitational radius
        self.R_g = (G * self.M_BH / np.power(c, 2)).to("cm")
        if R_g_units:
            # check that numbers have been passed
            R_in_unit_check = isinstance(R_in, (int, float))
            R_out_unit_check = isinstance(R_out, (int, float))
            if not R_in_unit_check or not R_out_unit_check:
                raise TypeError("R_in / R_out passed with units, int / float expected")
            self.R_in = R_in * self.R_g
            self.R_out = R_out * self.R_g
            self.R_in_tilde = R_in
            self.R_out_tilde = R_out
        else:
            # check that quantities have been passed
            R_in_unit_check = isinstance(R_in, u.Quantity)
            R_out_unit_check = isinstance(R_out, u.Quantity)
            if R_in_unit_check and R_out_unit_check:
                self.R_in = R_in
                self.R_out = R_out
                self.R_in_tilde = (self.R_in / self.R_g).to_value("")
                self.R_out_tilde = (self.R_out / self.R_g).to_value("")
            else:
                raise TypeError("R_in / R_out passed without units")
        # array of R_tile values
        self.R_tilde = np.linspace(self.R_in_tilde, self.R_out_tilde)

    def __str__(self):
        return (
            f"* Shakura Sunyaev accretion disk:\n"
            + f" - M_BH (central black hole mass): {self.M_BH.cgs:.2e}\n"
            + f" - L_disk (disk luminosity): {self.L_disk.cgs:.2e}\n"
            + f" - eta (accretion efficiency): {self.eta:.2e}\n"
            + f" - dot(m) (mass accretion rate): {self.m_dot.cgs:.2e}\n"
            + f" - R_in (disk inner radius): {self.R_in.cgs:.2e}\n"
            + f" - R_out (disk inner radius): {self.R_out.cgs:.2e}"
        )

    # staticmethods to be used in SED calculations without using a class instance
    @staticmethod
    def evaluate_mu_from_r_tilde(R_in_tilde, R_out_tilde, r_tilde, size=100):
        r"""array of cosine angles, spanning from :math:`R_{\mathrm{in}}` to 
        :math:`R_{\mathrm{out}}`, viewed from a given height :math:`\tilde{r}` 
        above the disk, Eq. 72 and 73 in [Finke2016]_."""
        mu_min = 1 / np.sqrt(1 + np.power((R_out_tilde / r_tilde), 2))
        mu_max = 1 / np.sqrt(1 + np.power((R_in_tilde / r_tilde), 2))
        return np.linspace(mu_min, mu_max, size)

    @staticmethod
    def evaluate_phi_disk_mu(mu, R_in_tilde, r_tilde):
        """dependency of the radiant surface-energy flux from the disk radius, 
        here obtained from the cosine of the zenith `mu` and the height above 
        the disk `r_tilde` (in graviational radius units), 
        Eq. 63 [Dermer2009]_"""
        R_tilde = r_tilde * np.sqrt(np.power(mu, -2) - 1)
        return 1 - np.sqrt(R_in_tilde / R_tilde)

    @staticmethod
    def evaluate_epsilon(L_disk, M_BH, eta, R_tilde):
        """evaluate the dimensionless energy emitted at the radius `R_tilde` 
        Eq. 65 [Dermer2009]_"""
        M_8 = (M_BH / (1e8 * M_sun)).to_value("")
        L_Edd = 1.26 * 1e46 * M_8 << u.Unit("erg s-1")
        l_Edd = (L_disk / L_Edd).to_value("")
        xi = np.power(l_Edd / (M_8 * eta), 1 / 4)
        return 2.7 * 1e-4 * xi * np.power(R_tilde, -3 / 4)

    @staticmethod
    def evaluate_epsilon_mu(L_disk, M_BH, eta, mu, r_tilde):
        """same as :func:`~agnpy.targets.SSDisk.evaluate_epsilon` but 
        considering the cosine of the subtended zenith `mu` and the height 
        above the disk `r` instead of the radius `R_tilde`"""
        R_tilde = r_tilde * np.sqrt(np.power(mu, -2) - 1)
        return SSDisk.evaluate_epsilon(L_disk, M_BH, eta, R_tilde)

    @staticmethod
    def evaluate_T(R, M_BH, m_dot, R_in):
        """black body temperature (K) at the radial coordinate R"""
        phi = 1 - np.sqrt((R_in / R).to_value(""))
        val = (3 * G * M_BH * m_dot * phi) / (8 * np.pi * np.power(R, 3) * sigma_sb)
        return np.power(val, 1 / 4).to("K")

    def epsilon_mu(self, mu, r_tilde):
        return self.evaluate_epsilon_mu(self.L_disk, self.M_BH, self.eta, mu, r_tilde)

    def epsilon(self, R_tilde):
        r"""dimensionless energy emitted at the disk radius :math:`\tilde{R}`"""
        return self.evaluate_epsilon(self.L_disk, self.M_BH, self.eta, R_tilde)

    def phi_disk_mu(self, mu, r_tilde):
        return self.evaluate_phi_disk_mu(mu, self.R_in_tilde, r_tilde)

    def phi_disk(self, R_tilde):
        return 1 - np.sqrt(self.R_in_tilde / R_tilde)

    def T(self, R):
        r"""temperature of the disk at radius :math:`R`. 
        Eq. 64 in [Dermer2009]_."""
        return self.evaluate_T(R, self.M_BH, self.m_dot, self.R_in)

    @staticmethod
    def evaluate_multi_T_bb_sed(nu, z, M_BH, m_dot, R_in, R_out, d_L, mu_s=1):
        r"""Evaluate a multi-temperature black body SED in the case of the SS Disk.
        The SED is calculated for an observer far away from the disk (:math:`d_L \gg R`) 
        with the following:

        .. math::
            \nu F_{\nu} \approx \mu_s \, \nu \frac{2 \pi}{d_L^2} 
            \int_{R_{\rm in}}^{R_{\rm out}}{\rm d}R \, R \, I_{\nu}(T(R)),

        where :math:`I_{\nu}` is Planck's law, :math:`R` the radial coordinate 
        along the disk, and :math:`d_L` the luminosity distance. :math:`\mu_s` 
        is the cosine of the angle between the disk axis and the observer's line of sight.

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        M_BH : :class:`~astropy.units.Quantity`
            Black Hole mass    
        m_dot : float
            mass accretion rate
        R_in : :class:`~astropy.units.Quantity`
            inner disk radius
        R_out : :class:`~astropy.units.Quantity`
            outer disk radius
        d_L : :class:`~astropy.units.Quantity` 
            luminosity of the source
        mu_s : float
            cosine of the angle between the observer line of sight and the disk axis
        """
        # correct for redshift
        nu *= 1 + z
        # array to integrate R
        R = np.linspace(R_in, R_out, 100)
        _R, _nu = axes_reshaper(R, nu)
        _T = SSDisk.evaluate_T(_R, M_BH, m_dot, R_in)
        _I_nu = BlackBody().evaluate(_nu, _T, scale=1)
        integrand = _R / np.power(d_L, 2) * _I_nu * u.sr
        F_nu = 2 * np.pi * np.trapz(integrand, R, axis=0)
        return mu_s * (nu * F_nu).to("erg cm-2 s-1")

    @staticmethod
    def evaluate_multi_T_bb_norm_sed(
        nu, z, L_disk, M_BH, m_dot, R_in, R_out, d_L, mu_s=1
    ):
        """Evaluate a normalised, multi-temperature black body SED as in
        :func:`~targets.SSDisk.evaluate_multi_T_bb_sed`, but the integral luminosity 
        is set to be equal to `L_disk`."""
        sed_disk = SSDisk.evaluate_multi_T_bb_sed(
            nu, z, M_BH, m_dot, R_in, R_out, d_L, mu_s
        )
        # renormalise, the factor 2 includes the two sides of the Disk
        L = 2 * (np.trapz(sed_disk / nu, nu) * 4 * np.pi * d_L ** 2).to("erg s-1")
        norm = (L_disk / L).to_value("")
        return norm * sed_disk

    def sed_flux(self, nu, z, mu_s=1):
        r"""evaluate the multi-temperature black body SED for this SS Disk, refer
        to :func:`~targets.SSDisk.evaluate_multi_T_bb_sed` and to 
        :func:`~targets.SSDisk.evaluate_multi_T_bb_norm_sed`
        """
        d_L = Distance(z=z).to("cm")
        return SSDisk.evaluate_multi_T_bb_norm_sed(
            nu, z, self.L_disk, self.M_BH, self.m_dot, self.R_in, self.R_out, d_L, mu_s
        )

    def u(self, r, blob=None):
        """integral energy density of radiation produced by the Disk at the distance 
        r along the jet axis. Integral over the solid angle of Eq. 69 in [Dermer2009]_.
        
        Parameters
        ----------
        r : :class:`~astropy.units.Quantity`
            array of distances along the jet axis
        blob : :class:`~agnpy.emission_regions.Blob`
            if provided, the energy density is computed in a reference frame 
            comvoing with the blob
        """
        r_tilde = (r / self.R_g).to_value("")
        mu = self.evaluate_mu_from_r_tilde(self.R_in_tilde, self.R_out_tilde, r_tilde)
        prefactor = (
            3
            * self.l_Edd
            * self.L_Edd
            * self.R_g
            / (8 * np.pi * c * self.eta * np.power(r, 3))
        )
        integrand = (
            1
            / mu
            * np.power(np.power(mu, -2) - 1, -3 / 2)
            * self.evaluate_phi_disk_mu(mu, self.R_in_tilde, r_tilde)
        )
        if blob:
            mu_prime = (mu - blob.Beta) / (1 - blob.Beta * mu)
            integrand_prefactor = 1 / (
                np.power(blob.Gamma, 6)
                * np.power(1 - blob.Beta * mu, 2)
                * np.power(1 + blob.Beta * mu_prime, 4)
            )
            integrand *= integrand_prefactor

        integral = np.trapz(integrand, mu, axis=0)
        return (prefactor * integral).to("erg cm-3")


class SphericalShellBLR:
    """Spherical Shell Broad Line Region, from [Finke2016]_.
    Each line is emitted from an infinitesimally thin spherical shell. 

    Parameters
    ----------
    L_disk : :class:`~astropy.units.Quantity`
        Luminosity of the disk whose radiation is being reprocessed by the BLR
    xi_line : float
        fraction of the disk radiation reprocessed by the BLR
    line : string
        type of line emitted
    R_line : :class:`~astropy.units.Quantity`
        radius of the BLR spherical shell
    """

    def __init__(self, L_disk, xi_line, line, R_line):
        self.name = "SphericalShellBLR"
        self.L_disk = L_disk
        self.xi_line = xi_line
        if line not in lines_dictionary:
            raise NameError(f"{line} not available in the line dictionary")
        self.line = line
        self.lambda_line = lines_dictionary[line]["lambda"]
        self.epsilon_line = (
            self.lambda_line.to("erg", equivalencies=u.spectral()) / mec2
        ).to_value("")
        self.R_line = R_line

    def __str__(self):
        return (
            f"* Spherical Shell Broad Line Region:\n"
            + f" - L_disk (accretion disk luminosity): {self.L_disk.cgs:.2e}\n"
            + f" - xi_line (fraction of the disk radiation reprocessed by the BLR): {self.xi_line:.2e}\n"
            + f" - line (type of emitted line): {self.line}, lambda = {self.lambda_line.cgs:.2e}\n"
            + f" - R_line (radius of the BLR shell): {self.R_line.cgs:.2e}\n"
        )

    def print_lines_list():
        r"""Print the list of the available spectral lines.
        The dictionary with the possible emission lines is taken from Table 5 in 
        [Finke2016]_ and contains the value of the line wavelength and the ratio of 
        its radius to the radius of the :math:`H_{\beta}` shell, not used at the moment.
        """
        for line in lines_dictionary.keys():
            print(f"{line}: {lines_dictionary[line]}")

    def u(self, r, blob=None):
        """Density of radiation produced by the BLR at the distance r along the 
        jet axis. Integral over the solid angle of Eq. 80 in [Finke2016]_.

        Parameters
        ----------
        r : :class:`~astropy.units.Quantity`
            array of distances along the jet axis
        blob : :class:`~agnpy.emission_regions.Blob`
            if provided, the energy density is computed in a reference frame 
            comvoing with the blob
        """
        mu = np.linspace(-1, 1)
        _mu = mu.reshape(mu.size, 1)
        _r = r.reshape(1, r.size)
        _x2 = np.power(_r, 2) + np.power(self.R_line, 2) - 2 * _r * self.R_line * _mu
        prefactor = self.xi_line * self.L_disk / (8 * np.pi * c)
        integrand = 1 / _x2
        if blob:
            _mu_star = np.sqrt(
                1 - np.power(self.R_line, 2) / _x2 * (1 - np.power(_mu, 2))
            )
            integrand_prefactor = np.power(blob.Gamma, 2) * np.power(
                1 - blob.Beta * _mu_star, 2
            )
            integrand *= integrand_prefactor
        integral = np.trapz(integrand, mu, axis=0)
        return (prefactor * integral).to("erg cm-3")


class RingDustTorus:
    """Dust Torus as infinitesimally thin annulus, from [Finke2016]_.
    For the Compton scattering monochromatic emission at the peak energy of the 
    Black Body spectrum is considered.

    Parameters
    ----------
    L_disk : :class:`~astropy.units.Quantity`
       Luminosity of the disk whose radiation is being reprocessed by the Torus
    xi_dt : float
        fraction of the disk radiation reprocessed
    T_dt : :class:`~astropy.units.Quantity`
        peak temperature of the black body emission of the Torus
    R_dt : :class:`~astropy.units.Quantity`
        radius of the Torus, if not specified the saturation radius of Eq. 96 in
        [Finke2016]_ will be used
    """

    def __init__(self, L_disk, xi_dt, T_dt, R_dt=None):
        self.name = "RingDustTorus"
        self.L_disk = L_disk
        self.xi_dt = xi_dt
        self.T_dt = T_dt
        # dimensionless temperature of the torus
        self.Theta = (k_B * self.T_dt / mec2).to_value("")
        self.epsilon_dt = 2.7 * self.Theta

        # if the radius is not specified use saturation radius Eq. 96 of [Finke2016]_
        if R_dt is None:
            self.R_dt = (
                3.5
                * 1e18
                * np.sqrt((self.L_disk / (1e45 * u.Unit("erg s-1"))).to_value(""))
                * np.power((self.T_dt / (1e3 * u.K)).to_value(""), -2.6)
            ) * u.cm
        else:
            self.R_dt = R_dt

    def __str__(self):
        return (
            f"* Ring Dust Torus:\n"
            + f" - L_disk (accretion disk luminosity): {self.L_disk.cgs:.2e}\n"
            + f" - xi_dt (fraction of the disk radiation reprocessed by the torus): {self.xi_dt:.2e}\n"
            + f" - T_dt (temperature of the dust torus): {self.T_dt:.2e}\n"
            + f" - R_dt (radius of the torus): {self.R_dt.cgs:.2e}\n"
        )

    @staticmethod
    def evaluate_bb_sed(nu, z, T_dt, R_dt, d_L):
        """evaluate the black body SED corresponding to the torus temperature"""
        nu *= 1 + z
        # geometrical factor for a source of size R_dt at distance d_L
        prefactor = np.pi * np.power((R_dt / d_L).to_value(""), 2) * u.sr
        I_nu = BlackBody().evaluate(nu, T_dt, scale=1)
        return (prefactor * nu * I_nu).to("erg cm-2 s-1")

    @staticmethod
    def evaluate_bb_norm_sed(nu, z, L_dt, T_dt, R_dt, d_L):
        """evaluate the torus black-body SED such that its integral luminosity 
        is equal to the torus luminosity (`xi_dt * L_disk`)"""
        sed_dt = RingDustTorus.evaluate_bb_sed(nu, z, T_dt, R_dt, d_L)
        # renormalise
        L = (np.trapz(sed_dt / nu, nu) * 4 * np.pi * d_L ** 2).to("erg s-1")
        norm = (L_dt / L).to_value("")
        return norm * sed_dt

    def sed_flux(self, nu, z):
        """evaluate the black-body SED for the Dust Torus"""
        d_L = Distance(z=z).to("cm")
        L_dt = self.xi_dt * self.L_disk
        return RingDustTorus.evaluate_bb_norm_sed(
            nu, z, L_dt, self.T_dt, self.R_dt, d_L
        )

    def u(self, r, blob=None):
        r"""Density of radiation produced by the Torus at the distance r along the 
        jet axis. Integral over the solid angle of Eq. 85 in [Finke2016]_

        Parameters
        ----------
        r : :class:`~astropy.units.Quantity`
            array of distances along the jet axis
        blob : :class:`~agnpy.emission_regions.Blob`
            if provided, the energy density is computed in a reference frame 
            comvoing with the blob
        """
        x2 = np.power(self.R_dt, 2) + np.power(r, 2)
        x = np.sqrt(x2)
        integral = self.xi_dt * self.L_disk / (4 * np.pi * c * x2)
        if blob:
            mu = (r / x).to_value("")
            integral *= np.power(blob.Gamma * (1 - blob.Beta * mu), 2)
        return integral.to("erg cm-3")



# module containing the synchrotron radiative process
import numpy as np
import astropy.units as u
from astropy.constants import e, h, c, m_e, sigma_T
from ..utils.math import axes_reshaper, gamma_to_integrate
from ..utils.conversion import nu_to_epsilon_prime, B_to_cgs, lambda_c


__all__ = ["R", "nu_synch_peak", "epsilon_B", "Synchrotron"]

e = e.gauss
B_cr = 4.414e13 * u.G  # critical magnetic field


def R(x):
    """Eq. 7.45 in [Dermer2009]_, angle-averaged integrand of the radiated power, the 
    approximation of this function, given in Eq. D7 of [Aharonian2010]_, is used.
    """
    term_1_num = 1.808 * np.power(x, 1 / 3)
    term_1_denom = np.sqrt(1 + 3.4 * np.power(x, 2 / 3))
    term_2_num = 1 + 2.21 * np.power(x, 2 / 3) + 0.347 * np.power(x, 4 / 3)
    term_2_denom = 1 + 1.353 * np.power(x, 2 / 3) + 0.217 * np.power(x, 4 / 3)
    return term_1_num / term_1_denom * term_2_num / term_2_denom * np.exp(-x)


def nu_synch_peak(B, gamma):
    """observed peak frequency for monoenergetic electrons 
    Eq. 7.19 in [DermerMenon2009]_"""
    B = B_to_cgs(B)
    nu_peak = (e * B / (2 * np.pi * m_e * c)) * np.power(gamma, 2)
    return nu_peak.to("Hz")


def calc_x(B_cgs, epsilon, gamma):
    """ratio of the frequency to the critical synchrotron frequency from 
    Eq. 7.34 in [DermerMenon2009]_, argument of R(x),
    note B has to be in cgs Gauss units"""
    x = (
        4
        * np.pi
        * epsilon
        * np.power(m_e, 2)
        * np.power(c, 3)
        / (3 * e * B_cgs * h * np.power(gamma, 2))
    )
    return x.to_value("")


def epsilon_B(B):
    r""":math:`\epsilon_B`, Eq. 7.21 [DermerMenon2009]_"""
    return (B / B_cr).to_value("")


def single_electron_synch_power(B_cgs, epsilon, gamma):
    """angle-averaged synchrotron power for a single electron, 
    to be folded with the electron distribution    
    """
    x = calc_x(B_cgs, epsilon, gamma)
    prefactor = np.sqrt(3) * np.power(e, 3) * B_cgs / h
    return prefactor * R(x)


def tau_to_attenuation(tau):
    """Converts the synchrotron self-absorption optical depth to an attenuation
    Eq. 7.122 in [DermerMenon2009]_."""
    u = 1 / 2 + np.exp(-tau) / tau - (1 - np.exp(-tau)) / np.power(tau, 2)
    return np.where(tau < 1e-3, 1, 3 * u / tau)


class Synchrotron:
    """Class for synchrotron radiation computation

    Parameters
    ----------
    blob : :class:`~agnpy.emission_region.Blob`
        emitting region and electron distribution 
    ssa : bool
        whether or not to consider synchrotron self absorption (SSA).    
        The absorption factor will be taken into account in
        :func:`~agnpy.synchrotron.Synchrotron.com_sed_emissivity`, in order to be
        propagated to :func:`~agnpy.synchrotron.Synchrotron.sed_luminosity` and
        :func:`~agnpy.synchrotron.Synchrotron.sed_flux`.
    integrator : func
        function to be used for integration (default = `np.trapz`)
	"""

    def __init__(self, blob, ssa=False, integrator=np.trapz):
        self.blob = blob
        self.ssa = ssa
        self.integrator = integrator

    @staticmethod
    def evaluate_tau_ssa(
        nu,
        z,
        d_L,
        delta_D,
        B,
        R_b,
        n_e,
        *args,
        integrator=np.trapz,
        gamma=gamma_to_integrate,
    ):
        """Computes the syncrotron self-absorption opacity for a general set
        of model parameters, see :func:`~agnpy:sycnhrotron.Synchrotron.evaluate_sed_flux`
        for parameters defintion. Eq. before 7.122 in [DermerMenon2009]_."""
        # conversions
        epsilon = nu_to_epsilon_prime(nu, z, delta_D)
        B_cgs = B_to_cgs(B)
        # multidimensional integration
        _gamma, _epsilon = axes_reshaper(gamma, epsilon)
        SSA_integrand = n_e.evaluate_SSA_integrand(_gamma, *args)
        integrand = SSA_integrand * single_electron_synch_power(B_cgs, _epsilon, _gamma)
        integral = integrator(integrand, gamma, axis=0)
        prefactor_k_epsilon = (
            -1 / (8 * np.pi * m_e * np.power(epsilon, 2)) * np.power(lambda_c / c, 3)
        )
        k_epsilon = (prefactor_k_epsilon * integral).to("cm-1")
        return (2 * k_epsilon * R_b).to_value("")

    @staticmethod
    def evaluate_sed_flux(
        nu,
        z,
        d_L,
        delta_D,
        B,
        R_b,
        n_e,
        *args,
        ssa=False,
        integrator=np.trapz,
        gamma=gamma_to_integrate,
    ):
        r"""Evaluates the flux SED (:math:`\nu F_{\nu}`) due to synchrotron radiation,
        for a general set of model parameters. Eq. 21 in [Finke2008]_.

        **Note** parameters after \*args need to be passed with a keyword

        Parameters
        ----------
        nu : :class:`~astropy.units.Quantity`
            array of frequencies, in Hz, to compute the sed 
            **note** these are observed frequencies (observer frame)
        z : float
            redshift of the source
        d_L : :class:`~astropy.units.Quantity` 
            luminosity distance of the source
        delta_D : float
            Doppler factor of the relativistic outflow
        B : :class:`~astropy.units.Quantity`
            magnetic field in the blob 
        R_b : :class:`~astropy.units.Quantity`
            size of the emitting region (spherical blob assumed)
        n_e : :class:`~agnpy.spectra.ElectronDistribution`
            electron energy distribution
        \*args
            parameters of the electron energy distribution (k_e, p, ...)
        ssa : bool
            whether to consider or not the self-absorption, default false
        integrator : func
            which function to use for integration, default `numpy.trapz`
        gamma : :class:`~numpy.ndarray`
            array of Lorentz factor over which to integrate the electron 
            distribution
        
        Returns
        -------
        :class:`~astropy.units.Quantity`
            array of the SED values corresponding to each frequency
        """
        # conversions
        epsilon = nu_to_epsilon_prime(nu, z, delta_D)
        B_cgs = B_to_cgs(B)
        # reshape for multidimensional integration
        _gamma, _epsilon = axes_reshaper(gamma, epsilon)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(_gamma, *args)
        # fold the electron distribution with the synchrotron power
        integrand = N_e * single_electron_synch_power(B_cgs, _epsilon, _gamma)
        emissivity = integrator(integrand, gamma, axis=0)
        prefactor = np.power(delta_D, 4) / (4 * np.pi * np.power(d_L, 2))
        sed = (prefactor * epsilon * emissivity).to("erg cm-2 s-1")

        if ssa:
            tau = Synchrotron.evaluate_tau_ssa(
                nu,
                z,
                d_L,
                delta_D,
                B,
                R_b,
                n_e,
                *args,
                integrator=integrator,
                gamma=gamma,
            )
            attenuation = tau_to_attenuation(tau)
            sed *= attenuation

        return sed

    @staticmethod
    def evaluate_sed_flux_delta_approx(nu, z, d_L, delta_D, B, R_b, n_e, *args):
        """Synchrotron flux SED using the delta approximation for the 
        synchrotron radiation Eq. 7.70 [DermerMenon2009]_."""
        epsilon_prime = nu_to_epsilon_prime(nu, z, delta_D)
        gamma_s = np.sqrt(epsilon_prime / epsilon_B(B))
        B_cgs = B_to_cgs(B)
        U_B = np.power(B_cgs, 2) / (8 * np.pi)
        V_b = 4 / 3 * np.pi * np.power(R_b, 3)
        N_e = V_b * n_e.evaluate(gamma_s, *args)
        prefactor = (
            np.power(delta_D, 4) * c * sigma_T * U_B / (6 * np.pi * np.power(d_L, 2))
        )
        value = prefactor * np.power(gamma_s, 3) * N_e
        return value.to("erg cm-2 s-1")

    def sed_flux(self, nu):
        r"""Evaluates the synchrotron flux SED for a Synchrotron object built 
        from a Blob."""
        return self.evaluate_sed_flux(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.B,
            self.blob.R_b,
            self.blob.n_e,
            *self.blob.n_e.parameters,
            ssa=self.ssa,
            integrator=self.integrator,
            gamma=self.blob.gamma,
        )

    def sed_flux_delta_approx(self, nu):
        """Evaluates the synchrotron flux SED using the delta approximation for 
        a Synchrotron object built from a blob."""
        return self.evaluate_sed_flux_delta_approx(
            nu,
            self.blob.z,
            self.blob.d_L,
            self.blob.delta_D,
            self.blob.B,
            self.blob.R_b,
            self.blob.n_e,
            *self.blob.n_e.parameters,
        )

    def sed_luminosity(self, nu):
        r"""Evaluates the synchrotron luminosity SED
        :math:`\nu L_{\nu} \, [\mathrm{erg}\,\mathrm{s}^{-1}]`
        for a a Synchrotron object built from a blob."""
        sphere = 4 * np.pi * np.power(self.blob.d_L, 2)
        return (sphere * self.sed_flux(nu)).to("erg s-1")

    def sed_peak_flux(self, nu):
        """provided a grid of frequencies nu, returns the peak flux of the SED
        """
        return self.sed_flux(nu).max()

    def sed_peak_nu(self, nu):
        """provided a grid of frequencies nu, returns the frequency at which the 
        SED peaks
        """
        idx_max = self.sed_flux(nu).argmax()
        return nu[idx_max]



# conversion utilities for agnpy
from astropy.constants import m_e, h, c, G
import astropy.units as u


mec2 = m_e.to("erg", equivalencies=u.mass_energy())
lambda_c = (h / (m_e * c)).to("cm")  # Compton wavelength
# equivalency for decomposing Gauss in Gaussian-cgs units (not available in astropy)
Gauss_cgs_unit = "cm(-1/2) g(1/2) s-1"
Gauss_cgs_equivalency = [(u.G, u.Unit(Gauss_cgs_unit), lambda x: x, lambda x: x)]
# equivalency to transform frequencies to energies in electron rest mass units
epsilon_equivalency = [
    (u.Hz, u.Unit(""), lambda x: h.cgs * x / mec2, lambda x: x * mec2 / h.cgs)
]


def nu_to_epsilon_prime(nu, z=0, delta_D=1):
    """convert the frequency to a dimensionless energy in another reference
    frame with redshift z and moving with doppler factor delta_D"""
    epsilon = nu.to("", equivalencies=epsilon_equivalency)
    return (1 + z) * epsilon / delta_D


def B_to_cgs(B):
    """convert a magnetic field to CGS units"""
    return B.to(Gauss_cgs_unit, equivalencies=Gauss_cgs_equivalency)


def to_R_g_units(r, M):
    """convert a distance in graviational radii units"""
    R_g = G * M / c ** 2
    return (r / R_g).to("")



# tests on targets module
import numpy as np
import astropy.units as u
from astropy.coordinates import Distance
from astropy.constants import e, c, m_e, M_sun, G, sigma_sb
from agnpy.emission_regions import Blob
from agnpy.targets import (
    CMB,
    PointSourceBehindJet,
    SSDisk,
    SphericalShellBLR,
    RingDustTorus,
)
import pytest

# variables with _test are global and meant to be used in all tests
blob_test = Blob()

# CMB at z = 1
cmb_test = CMB(z=1)

# PointSourceBehindJet
ps_test = PointSourceBehindJet(1e46 * u.Unit("erg s-1"), 1e-5)

# disk
M_BH_test = 1.2 * 1e9 * M_sun
L_disk_test = 1.512 * 1e46 * u.Unit("erg s-1")
eta_test = 1 / 12
R_g_test = 1.77 * 1e14 * u.cm
R_in_tilde_test = 6
R_out_tilde_test = 200
R_in_test = R_in_tilde_test * R_g_test
R_out_test = R_out_tilde_test * R_g_test
disk_test = SSDisk(M_BH_test, L_disk_test, eta_test, R_in_test, R_out_test)
# useful for checks
L_Edd_test = 15.12 * 1e46 * u.Unit("erg s-1")
m_dot_test = 2.019 * 1e26 * u.Unit("g s-1")

# SphericalShellBLR
blr_test = SphericalShellBLR(L_disk_test, 0.1, "Lyalpha", 1e17 * u.cm)

# dust torus definition
dt_test = RingDustTorus(L_disk_test, 0.1, 1000 * u.K)


class TestCMB:
    """class grouping all the tests related to the CMB"""

    def test_u(self):
        """test u in the stationary reference frame"""
        assert u.isclose(
            6.67945605e-12 * u.Unit("erg / cm3"),
            cmb_test.u(),
            atol=0 * u.Unit("erg / cm3"),
            rtol=1e-5,
        )

    def test_u_comoving(self):
        """test u in the reference frame comoving with the blob"""
        assert u.isclose(
            8.88373221e-10 * u.Unit("erg / cm3"),
            cmb_test.u(blob_test),
            atol=0 * u.Unit("erg / cm3"),
            rtol=1e-3,
        )


class TestPointSourceBehindJet:
    """class grouping all the tests related to the PointSourceBehindJet"""

    def test_u(self):
        """test u in the stationary reference frame"""
        r = np.asarray([1e18, 1e19, 1e20]) * u.cm
        assert u.allclose(
            [2.65441873e-02, 2.65441873e-04, 2.65441873e-06] * u.Unit("erg / cm3"),
            ps_test.u(r),
            atol=0 * u.Unit("erg / cm3"),
            rtol=1e-3,
        )

    def test_u_comoving(self):
        """test u in the reference frame comoving with the blob"""
        r = np.asarray([1e18, 1e19, 1e20]) * u.cm
        assert u.allclose(
            [6.6693519e-05, 6.6693519e-07, 6.6693519e-09] * u.Unit("erg / cm3"),
            ps_test.u(r, blob_test),
            atol=0 * u.Unit("erg / cm3"),
            rtol=1e-3,
        )


class TestSSDisk:
    """class grouping all the tests related to the SSDisk target"""

    # global quantities defining the test disk
    def test_L_Edd(self):
        assert u.isclose(
            disk_test.L_Edd, L_Edd_test, atol=0 * u.Unit("erg s-1"), rtol=1e-3
        )

    def test_l_Edd(self):
        assert u.isclose(disk_test.l_Edd, 0.1, atol=0, rtol=1e-3)

    def test_m_dot(self):
        assert u.isclose(
            disk_test.m_dot, m_dot_test, atol=0 * u.Unit("g s-1"), rtol=1e-3
        )

    @pytest.mark.parametrize(
        "R_in, R_out, R_g_units",
        [(R_in_tilde_test, R_out_tilde_test, False), (R_in_test, R_out_test, True),],
    )
    def test_R_in_R_out_units(self, R_in, R_out, R_g_units):
        """check if a TypeError is raised when passing R_in and R_out with 
        (without) units but specifiying R_g_units True (False)"""
        with pytest.raises(TypeError):
            disk = SSDisk(M_BH_test, L_disk_test, eta_test, R_in, R_out, R_g_units)

    def test_R_g(self):
        assert u.isclose(disk_test.R_g, R_g_test, atol=0 * u.cm, rtol=1e-2)

    def test_mu_from_r_tilde(self):
        mu = SSDisk.evaluate_mu_from_r_tilde(
            R_in_tilde_test, R_out_tilde_test, r_tilde=10
        )
        mu_min_expected = 0.050
        mu_max_expected = 0.858
        assert np.isclose(mu[0], mu_min_expected, atol=0, rtol=1e-2)
        assert np.isclose(mu[-1], mu_max_expected, atol=0, rtol=1e-2)

    def test_phi_disk(self):
        R_tilde = 10
        phi_expected = 0.225
        assert np.isclose(disk_test.phi_disk(R_tilde), phi_expected, atol=0, rtol=1e-2)

    def test_phi_disk_mu(self):
        r_tilde = 10
        mu = 1 / np.sqrt(2)
        phi_disk_expected = 0.225
        phi_disk = SSDisk.evaluate_phi_disk_mu(mu, R_in_tilde_test, r_tilde)
        assert np.allclose(phi_disk, phi_disk_expected, atol=0, rtol=1e-2)

    def test_epsilon(self):
        R_tilde = 10
        epsilon_expected = 2.7e-5
        assert np.allclose(
            disk_test.epsilon(R_tilde), epsilon_expected, atol=0, rtol=1e-2
        )

    def test_epsilon_mu(self):
        r_tilde = 10
        # assume R_tilde = 10 as before
        mu = 1 / np.sqrt(2)
        epsilon_disk_expected = 2.7e-5
        epsilon_disk = SSDisk.evaluate_epsilon_mu(
            L_disk_test, M_BH_test, eta_test, mu, r_tilde
        )
        assert np.allclose(epsilon_disk, epsilon_disk_expected, atol=0, rtol=1e-2)

    def test_T(self):
        R = 10 * R_g_test
        phi = 1 - np.sqrt((disk_test.R_in / R).to(""))
        T_expected = np.power(
            3
            * G
            * M_BH_test
            * m_dot_test
            * phi
            / (8 * np.pi * np.power(R, 3) * sigma_sb),
            1 / 4,
        ).to("K")
        assert u.isclose(disk_test.T(R), T_expected, atol=0 * u.K, rtol=1e-2)

    @pytest.mark.parametrize("R_out", [1e2, 1e3, 1e4])
    @pytest.mark.parametrize("mu_s", [0.5, 0.8, 1.0])
    def test_bb_sed_luminosity(self, R_out, mu_s):
        """test that the luminosity of the disk BB SED is the same as L_disk,
        create disks with different outer radii"""
        disk = SSDisk(M_BH_test, L_disk_test, 1 / 12, 6, R_out, R_g_units=True)
        # compute the SEDs, assume a random redshift
        z = 0.23
        nu = np.logspace(10, 20, 100) * u.Hz
        sed = disk.sed_flux(nu, z, mu_s)
        # compute back the luminosity
        d_L = Distance(z=z).to("cm")
        F_nu = sed / nu
        # renormalise, the factor 2 includes the two sides of the Disk
        L = 2 * (4 * np.pi * np.power(d_L, 2) * np.trapz(F_nu, nu, axis=0))
        assert u.isclose(L, L_disk_test, atol=0 * u.Unit("erg s-1"), rtol=1e-2)


class TestSphericalShellBLR:
    """class grouping all the tests related to the SphericalShellBLR target"""

    @pytest.mark.parametrize(
        "line, lambda_line",
        [
            ("Lyalpha", 1215.67 * u.Angstrom),
            ("Lybeta", 1025.72 * u.Angstrom),
            ("Halpha", 6564.61 * u.Angstrom),
            ("Hbeta", 4862.68 * u.Angstrom),
        ],
    )
    def test_line_dict(self, line, lambda_line):
        """test correct loading of some of the emission line"""
        blr = SphericalShellBLR(1e46 * u.Unit("erg s-1"), 0.1, line, 1e17)
        assert u.isclose(blr.lambda_line, lambda_line, atol=0 * u.Angstrom)

    def test_u(self):
        """test u in the stationary reference frame"""
        r = np.logspace(16, 20, 10) * u.cm
        assert u.allclose(
            [
                4.02698710e-01,
                4.12267268e-01,
                5.49297935e-01,
                9.36951182e-02,
                1.12734943e-02,
                1.44410780e-03,
                1.86318218e-04,
                2.40606696e-05,
                3.10750072e-06,
                4.01348246e-07,
            ]
            * u.Unit("erg / cm3"),
            blr_test.u(r),
            atol=0 * u.Unit("erg / cm3"),
        )

    def test_u_blr_vs_point_source(self):
        """test that for large enough distances the energy density of the 
        BLR tends to the one of a point like source approximating it"""
        # point source with the same luminosity as the BLR
        ps_blr = PointSourceBehindJet(
            blr_test.xi_line * blr_test.L_disk, blr_test.epsilon_line
        )
        # r >> R_line
        r = np.logspace(19, 23, 10) * u.cm
        assert u.allclose(
            blr_test.u(r), ps_blr.u(r), atol=0 * u.Unit("erg cm-3"), rtol=1e-2
        )

    def test_u_comoving(self):
        """test u in the reference frame comoving with the blob"""
        r = np.logspace(16, 20, 10) * u.cm
        assert u.allclose(
            [
                1.35145224e01,
                1.39362806e01,
                1.98574615e01,
                7.22733332e-02,
                2.50276530e-04,
                5.60160671e-06,
                4.97415343e-07,
                6.09348663e-08,
                7.81583912e-09,
                1.00855244e-09,
            ]
            * u.Unit("erg / cm3"),
            blr_test.u(r, blob_test),
            atol=0 * u.Unit("erg / cm3"),
        )

    def test_u_blr_vs_point_source_comoving(self):
        """test that for large enough distances the energy density of the 
        BLR tends to the one of a point like source approximating it"""
        # point source with the same luminosity as the BLR
        ps_blr = PointSourceBehindJet(
            blr_test.xi_line * blr_test.L_disk, blr_test.epsilon_line
        )
        # r >> R_line
        r = np.logspace(19, 23, 10) * u.cm
        assert u.allclose(
            blr_test.u(r, blob_test),
            ps_blr.u(r, blob_test),
            atol=0 * u.Unit("erg cm-3"),
            rtol=1e-1,
        )


class TestRingDustTorus:
    """class grouping all the tests related to the RingDustTorus target"""

    def test_sublimation_radius(self):
        assert u.allclose(dt_test.R_dt, 1.361 * 1e19 * u.cm, atol=0 * u.cm, rtol=1e-3)

    def test_setting_radius(self):
        """check that, when passed manually, the radius is correctly set"""
        dt = RingDustTorus(L_disk_test, 0.1, 1e3 * u.K, 1e19 * u.cm)
        assert u.allclose(dt.R_dt, 1e19 * u.cm, atol=0 * u.cm)

    @pytest.mark.parametrize("T_dt", [1e3, 2e3, 5e3] * u.K)
    def test_bb_sed_luminosity(self, T_dt):
        """test that the luminosity of the DT BB SED is the same as xi_dt * L_disk,
        create DTs with different temperatrues (and radii)"""
        xi_dt = 0.5
        L_dt = xi_dt * L_disk_test
        dt = RingDustTorus(L_disk_test, xi_dt, T_dt)
        # compute the SEDs, assume a random redshift
        z = 0.23
        nu = np.logspace(10, 20, 100) * u.Hz
        sed = dt.sed_flux(nu, z)
        # compute back the luminosity
        d_L = Distance(z=z).to("cm")
        F_nu = sed / nu
        L = 4 * np.pi * np.power(d_L, 2) * np.trapz(F_nu, nu, axis=0)
        assert u.isclose(L, L_dt, atol=0 * u.Unit("erg s-1"), rtol=1e-2)

    def test_u(self):
        """test u in the stationary reference frame"""
        r = np.logspace(17, 23, 10) * u.cm
        assert u.allclose(
            [
                2.16675545e-05,
                2.16435491e-05,
                2.11389842e-05,
                1.40715277e-05,
                1.71541601e-06,
                8.61241559e-08,
                4.01273788e-09,
                1.86287690e-10,
                8.64677950e-12,
                4.01348104e-13,
            ]
            * u.Unit("erg / cm3"),
            dt_test.u(r),
            atol=0 * u.Unit("erg / cm3"),
        )

    def test_u_dt_vs_point_source(self):
        """test that in the stationary reference frame, for large enough 
        distances, the energy density of the DT tends to the one of a point like 
        source approximating it"""
        # point source with the same luminosity as the DT
        ps_dt = PointSourceBehindJet(dt_test.xi_dt * dt_test.L_disk, dt_test.epsilon_dt)
        # r >> R_dt
        r = np.logspace(21, 24, 10) * u.cm
        assert u.allclose(
            dt_test.u(r), ps_dt.u(r), atol=0 * u.Unit("erg cm-3"), rtol=1e-2
        )

    def test_u_comoving(self):
        """test u in the reference frame comoving with the blob"""
        r = np.logspace(17, 23, 10) * u.cm
        assert u.allclose(
            [
                2.13519004e-03,
                2.02003750e-03,
                1.50733234e-03,
                2.37521472e-04,
                3.50603715e-07,
                4.21027697e-10,
                1.04563603e-11,
                4.68861573e-13,
                2.17274347e-14,
                1.00842240e-15,
            ]
            * u.Unit("erg / cm3"),
            dt_test.u(r, blob_test),
            atol=0 * u.Unit("erg / cm3"),
        )

    def test_u_dt_vs_point_source_comoving(self):
        """test that in the reference frame comoving with the Blob, for large 
        enough distances, the energy density of the DT tends to the one of 
        a point like source approximating it"""
        # point source with the same luminosity as the DT
        ps_dt = PointSourceBehindJet(dt_test.xi_dt * dt_test.L_disk, dt_test.epsilon_dt)
        # r >> R_line
        r = np.logspace(21, 24, 10) * u.cm
        assert u.allclose(
            dt_test.u(r, blob_test),
            ps_dt.u(r, blob_test),
            atol=0 * u.Unit("erg cm-3"),
            rtol=1e-1,
        )



# geometry utilities for agnpy
import numpy as np


def cos_psi(mu_s, mu, phi):
    """compute the angle between the blob (with zenith mu_s) and a photon with
    zenith and azimuth (mu, phi). The system is symmetric in azimuth for the
    electron phi_s = 0, Eq. 8 in [Finke2016]_."""
    term_1 = mu * mu_s
    term_2 = np.sqrt(1 - np.power(mu, 2)) * np.sqrt(1 - np.power(mu_s, 2))
    term_3 = np.cos(phi)
    return term_1 + term_2 * term_3


def x_re_shell(mu, R_re, r):
    """distance between the blob and a spherical reprocessing material,
    see Fig. 9 and Eq. 76 in [Finke2016]_.

    Parameters
    ----------
    mu : :class:`~numpy.ndarray`
        (array of) cosine of the zenith angle subtended by the target
    R_re : :class:`~astropy.units.Quantity`
        distance from the BH to the reprocessing material
    r : :class:`~astropy.units.Quantity`
        height of the emission region in the jet
    """
    return np.sqrt(np.power(R_re, 2) + np.power(r, 2) - 2 * r * R_re * mu)


def mu_star_shell(mu, R_re, r):
    """cosine of the angle between the blob and a sphere of reprocessing
    material, see Fig. 9 and Eq. 78 in [Finke2016]_.
    works only if the blob is at the jet axis

    Parameters
    ----------
    mu : :class:`~numpy.ndarray`
        (array of) cosine of the zenith angle subtended by the target
    R_re : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the reprocessing material
    r : :class:`~astropy.units.Quantity`
        height (in cm) of the emission region in the jet
    """
    addend = np.power(R_re / x_re_shell(mu, R_re, r), 2) * (1 - np.power(mu, 2))
    mu_star = np.sqrt(1 - addend)
    # if r < mu * R_re you need to multiply by -1 because the blob is before the shell element
    mu_sign = ((r > mu * R_re) - 0.5) * 2
    mu_star *= mu_sign
    return mu_star


def x_re_ring(R_re, r):
    """distance between the blob and a ring of reprocessing material"""
    return np.sqrt(np.power(R_re, 2) + np.power(r, 2))


def x_re_ring_mu_s(R_re, r, phi_re, u, mu_s):
    """distance between the blob and a ring of reprocessing material
    if the photon moved u along the mu_s direction starting from r position

    Parameters
    ----------
    R_re : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the reprocessing material
    r : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the starting place of the photon
    phi_re : :class:`~numpy.ndarray`
        (array of) azimuth angle of the reprocessing material 
    u : :class:`~numpy.ndarray`
        (array of) distances (in cm) that the gamma ray has travelled 
    mu_s : :class:`~astropy.units.Quantity`
        direction of gamma ray  motion: cos angle to the jet axis. 
        The gamma ray is moving in the direction of phi_re=0
    """

    sin_theta_s = np.sqrt(1 - mu_s * mu_s)
    x2 = (
        u * u
        + R_re * R_re
        + r * r
        - 2 * u * R_re * sin_theta_s * np.cos(phi_re)
        + 2 * r * u * mu_s
    )
    return np.sqrt(x2)


def phi_mu_re_ring(R_re, r, phi_re, u, mu_s):
    """azimuth and  angle of the soft photon produced from  DT/BLR ring
    photon is produced at (R_re*cos(phi_re), R_re*sin(phi_re),0)
    and reached gamma ray at (u*sin(theta_s), 0, r+u*mu_s) distance between the blob and a ring of reprocessing material
    if the photon moved u along the mu_s direction starting from r position

    Parameters
    ----------
    R_re : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the reprocessing material
    r : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the starting place of the photon
    phi_re : :class:`~numpy.ndarray`
        (array of) azimuth angle of the reprocessing material 
    u : :class:`~numpy.ndarray`
        (array of) distances (in cm) that the gamma ray has travelled 
    mu_s : :class:`~astropy.units.Quantity`
        direction of gamma ray  motion: cos angle to the jet axis. 
        The gamma ray is moving in the direction of phi_re=0
    """
    sin_theta_s = np.sqrt(1 - mu_s * mu_s)
    dx = u * sin_theta_s - R_re * np.cos(phi_re)
    dy = 0 - R_re * np.sin(phi_re)
    dz = r + u * mu_s
    phi = np.arctan2(dy, dx)
    x = x_re_ring_mu_s(R_re, r, phi_re, u, mu_s)
    mu = dz / x
    return phi, mu


def x_re_shell_mu_s(R_re, r, phi_re, mu_re, u, mu_s):
    """distance between the blob and a spherical reprocessing material,
    if the photon moved u along the mu_s direction starting from r position

    Parameters
    ----------
    R_re : :class:`~astropy.units.Quantity`
        distance from the BH to the reprocessing material
    r : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the starting place of the photon
    phi_re : :class:`~numpy.ndarray`
        (array of) azimuth angle of the reprocessing material 
    mu_re : :class:`~numpy.ndarray`
        (array of) cos of zenith angle of the reprocessing material 
    u : :class:`~numpy.ndarray`
        (array of) distances (in cm) that the gamma ray has travelled 
    mu_s : :class:`~astropy.units.Quantity`
        direction of gamma ray  motion: cos angle to the jet axis. 
        The gamma ray is moving in the direction of phi_re=0
    """

    sin_re = np.sqrt(1 - mu_re * mu_re)
    sin_s = np.sqrt(1 - mu_s * mu_s)
    x2 = (
        R_re * R_re
        + u * u
        + r * r
        - 2 * R_re * u * sin_re * np.cos(phi_re) * sin_s
        - 2 * R_re * mu_re * (r + u * mu_s)
        + 2 * r * u * mu_s
    )

    return np.sqrt(x2)


def phi_mu_re_shell(R_re, r, phi_re, mu_re, u, mu_s):
    """azimuth and  angle of the soft photon produced from  BLR shell
    photon is produced at 
    (R_re*sin(th_re)*cos(phi_re), R_re*sin(th_re)*sin(phi_re),R_re*cos(th_re))
    and reached gamma ray at (u*sin(theta_s), 0, r+u*mu_s) distance between the blob and a ring of reprocessing material
    if the photon moved u along the mu_s direction starting from r position

    Parameters
    ----------
    R_re : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the reprocessing material
    r : :class:`~astropy.units.Quantity`
        distance (in cm) from the BH to the starting place of the photon
    phi_re : :class:`~numpy.ndarray`
        (array of) azimuth angle of the reprocessing material 
    mu_re : :class:`~numpy.ndarray`
        (array of) cos of zenith angle of the reprocessing material 
    u : :class:`~numpy.ndarray`
        (array of) distances (in cm) that the gamma ray has travelled 
    mu_s : :class:`~astropy.units.Quantity`
        direction of gamma ray  motion: cos angle to the jet axis. 
        The gamma ray is moving in the direction of phi_re=0
    """
    sin_theta_s = np.sqrt(1 - mu_s * mu_s)
    sin_re = np.sqrt(1 - mu_re * mu_re)
    dx = u * sin_theta_s - R_re * sin_re * np.cos(phi_re)
    dy = -R_re * sin_re * np.sin(phi_re)
    dz = r + u * mu_s - R_re * mu_re
    phi = np.arctan2(dy, dx)
    x = x_re_shell_mu_s(R_re, r, phi_re, mu_re, u, mu_s)
    mu = dz / x
    return phi, mu



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="agnpy",
    version="0.1.1",
    author="Cosimo Nigro",
    author_email="cosimonigro2@gmail.com.com",
    description="Modelling jetted Active Galactic Nuclei radiative processes with python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cosimoNigro/agnpy",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=["astropy>=4.0", "numpy>=1.17", "scipy>=1.2", "matplotlib"],
    python_requires=">=3.6",
)



# tests on spectra module
import numpy as np
import astropy.units as u
from astropy.constants import m_e
from agnpy.spectra import PowerLaw, BrokenPowerLaw, LogParabola, ExpCutoffPowerLaw
from agnpy.utils.math import trapz_loglog
from agnpy.utils.conversion import mec2
import pytest

# variables with _test are global and meant to be used in all tests
# global PowerLaw
k_e_test = 1e-13 * u.Unit("cm-3")
p_test = 2.1
gamma_min_test = 10
gamma_max_test = 1e7
pwl_test = PowerLaw(k_e_test, p_test, gamma_min_test, gamma_max_test)
# global BrokenPowerLaw
p1_test = 2.1
p2_test = 3.1
gamma_b_test = 1e3
bpwl_test = BrokenPowerLaw(
    k_e_test, p1_test, p2_test, gamma_b_test, gamma_min_test, gamma_max_test
)
# global LogParabola
q_test = 0.2
gamma_0_test = 1e4
lp_test = LogParabola(
    k_e_test, p_test, q_test, gamma_0_test, gamma_min_test, gamma_max_test
)
# global PowerLaw exp Cutoff
gamma_c_test = 1e3
epwl_test = ExpCutoffPowerLaw(
    k_e_test, p_test, gamma_c_test, gamma_min_test, gamma_max_test
)


def power_law_integral(k_e, p, gamma_min, gamma_max):
    """analytical integral of the power law"""
    if np.isclose(p, 1.0):
        integral = np.log(gamma_max / gamma_min)
    else:
        integral = (np.power(gamma_max, 1 - p) - np.power(gamma_min, 1 - p)) / (1 - p)
    return k_e * integral


def power_law_times_gamma_integral(k_e, p, gamma_min, gamma_max):
    """analytical integral of the power law multiplied by gamma"""
    if np.isclose(p, 2.0):
        integral = np.log(gamma_max / gamma_min)
    else:
        integral = (np.power(gamma_max, 2 - p) - np.power(gamma_min, 2 - p)) / (2 - p)
    return k_e * integral


def broken_power_law_integral(k_e, p1, p2, gamma_b, gamma_min, gamma_max):
    """analytical integral of the power law with two spectral indexes"""
    if np.allclose(p1, 1.0):
        term_1 = gamma_b * np.log(gamma_b / gamma_min)
    else:
        term_1 = gamma_b * (1 - np.power(gamma_min / gamma_b, 1 - p1)) / (1 - p1)
    if np.allclose(p2, 1.0):
        term_2 = gamma_b * np.log(gamma_max / gamma_b)
    else:
        term_2 = gamma_b * (np.power(gamma_max / gamma_b, 1 - p2) - 1) / (1 - p2)
    return k_e * (term_1 + term_2)


def broken_power_law_times_gamma_integral(k_e, p1, p2, gamma_b, gamma_min, gamma_max):
    """analytical integral of the power law with two spectral indexes multiplied 
    by gamma"""
    if np.allclose(p1, 2.0):
        term_1 = np.power(gamma_b, 2) * np.log(gamma_b / gamma_min)
    else:
        term_1 = (
            np.power(gamma_b, 2)
            * (1 - np.power(gamma_min / gamma_b, 2 - p1))
            / (2 - p1)
        )
    if np.allclose(p2, 2.0):
        term_2 = np.power(gamma_b, 2) * np.log(gamma_max / gamma_b)
    else:
        term_2 = (
            np.power(gamma_b, 2)
            * (np.power(gamma_max / gamma_b, 2 - p2) - 1)
            / (2 - p2)
        )
    return k_e * (term_1 + term_2)


class TestPowerLaw:
    """class grouping all tests related to the BrokenPowerLaw spectrum"""

    def test_call(self):
        """assert that outside the bounding box the function returns 0"""
        gamma = np.logspace(0, 8)
        values = pwl_test(gamma).value
        condition = (gamma_min_test <= gamma) * (gamma <= gamma_max_test)
        # check that outside the boundaries values are all 0
        assert not np.all(values[~condition])

    @pytest.mark.parametrize("p", [1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0])
    @pytest.mark.parametrize("integrator", [np.trapz, trapz_loglog])
    def test_power_law_integral(self, p, integrator):
        """test the integration of the power law for different spectral indexes
        different integrating functions"""
        pwl = PowerLaw(
            k_e_test, p, gamma_min_test, gamma_max_test, integrator=integrator
        )
        numerical_integral = pwl.integral(gamma_min_test, gamma_max_test)
        analytical_integral = power_law_integral(
            k_e_test, p, gamma_min_test, gamma_max_test
        )
        assert u.isclose(
            numerical_integral, analytical_integral, atol=0 * u.Unit("cm-3"), rtol=1e-2
        )

    @pytest.mark.parametrize("p", [1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0])
    @pytest.mark.parametrize("integrator", [np.trapz, trapz_loglog])
    def test_power_law_times_gamma_integral(self, p, integrator):
        """test the integration of the power law times gamma for different 
        spectral indexes and different integrating functions"""
        pwl = PowerLaw(
            k_e_test, p, gamma_min_test, gamma_max_test, integrator=integrator
        )
        numerical_integral = pwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=1
        )
        analytical_integral = power_law_times_gamma_integral(
            k_e_test, p, gamma_min_test, gamma_max_test,
        )
        assert u.isclose(
            numerical_integral, analytical_integral, atol=0 * u.Unit("cm-3"), rtol=1e-2
        )

    def test_from_normalised_density(self):
        """test the intialisation of the power law from the total particle 
        density"""
        n_e_tot = 1e-5 * u.Unit("cm-3")
        pwl = PowerLaw.from_normalised_density(
            n_e_tot=n_e_tot,
            p=p_test,
            gamma_min=gamma_min_test,
            gamma_max=gamma_max_test,
        )
        # calculate n_e_tot
        n_e_tot_calc = pwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=0
        )
        assert u.isclose(n_e_tot, n_e_tot_calc, atol=0 * u.Unit("cm-3"), rtol=1e-2)

    def test_from_normalised_energy_density(self):
        """test the intialisation of the power law from the total particle 
        energy density"""
        u_e = 3e-4 * u.Unit("erg cm-3")
        pwl = PowerLaw.from_normalised_energy_density(
            u_e=u_e, p=p_test, gamma_min=gamma_min_test, gamma_max=gamma_max_test
        )
        # calculate u_e
        u_e_calc = mec2 * pwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=1
        )
        assert u.isclose(u_e, u_e_calc, atol=0 * u.Unit("erg cm-3"), rtol=1e-2)

    def test_from_norm_at_gamma_1(self):
        """test the intialisation of the powerlaw from the normalisation at 
        gamma = 1"""
        norm = 1e-13 * u.Unit("cm-3")
        pwl = PowerLaw.from_norm_at_gamma_1(
            norm=norm, p=p_test, gamma_min=1, gamma_max=gamma_max_test
        )
        assert u.isclose(norm, pwl(1), atol=0 * u.Unit("cm-3"), rtol=1e-2)


class TestBrokenPowerLaw:
    """class grouping all tests related to the BrokenPowerLaw spectrum"""

    def test_call(self):
        """assert that outside the bounding box the function returns 0"""
        gamma = np.logspace(0, 8)
        values = bpwl_test(gamma).value
        condition = (gamma_min_test <= gamma) * (gamma <= gamma_max_test)
        # check that outside the boundaries values are all 0
        assert not np.all(values[~condition])

    @pytest.mark.parametrize("p1", [1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0])
    @pytest.mark.parametrize("p2", [1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0])
    @pytest.mark.parametrize("gamma_b", [1e2, 1e3, 1e4, 1e5, 1e6])
    @pytest.mark.parametrize("integrator", [np.trapz, trapz_loglog])
    def test_broken_power_law_integral(self, p1, p2, gamma_b, integrator):
        """test the integration of the log parabola for different spectral 
        indexes and breaks and different integrating functions"""
        bpwl = BrokenPowerLaw(
            k_e_test,
            p1,
            p2,
            gamma_b,
            gamma_min_test,
            gamma_max_test,
            integrator=integrator,
        )
        numerical_integral = bpwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=0
        )
        analytical_integral = broken_power_law_integral(
            k_e_test, p1, p2, gamma_b, gamma_min_test, gamma_max_test
        )
        assert u.isclose(
            numerical_integral, analytical_integral, atol=0 * u.Unit("cm-3"), rtol=1e-2
        )

    @pytest.mark.parametrize("p1", [1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0])
    @pytest.mark.parametrize("p2", [1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0])
    @pytest.mark.parametrize("gamma_b", [1e2, 1e3, 1e4, 1e5, 1e6])
    @pytest.mark.parametrize("integrator", [np.trapz, trapz_loglog])
    def test_broken_power_law_times_gamma_integral(self, p1, p2, gamma_b, integrator):
        """test the integration of the broken power law times gamma for different 
        spectral indexes and breaks and different integrating functions"""
        bpwl = BrokenPowerLaw(
            k_e_test,
            p1,
            p2,
            gamma_b,
            gamma_min_test,
            gamma_max_test,
            integrator=integrator,
        )
        numerical_integral = bpwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=1
        )
        analytical_integral = broken_power_law_times_gamma_integral(
            k_e_test, p1, p2, gamma_b, gamma_min_test, gamma_max_test
        )
        assert u.isclose(
            numerical_integral, analytical_integral, atol=0 * u.Unit("cm-3"), rtol=1e-2
        )

    def test_from_normalised_density(self):
        """test the intialisation of the broken power law from the total particle 
        density"""
        n_e_tot = 1e-5 * u.Unit("cm-3")
        bpwl = BrokenPowerLaw.from_normalised_density(
            n_e_tot=n_e_tot,
            p1=p1_test,
            p2=p2_test,
            gamma_b=gamma_b_test,
            gamma_min=gamma_min_test,
            gamma_max=gamma_max_test,
        )
        # calculate n_e_tot
        n_e_tot_calc = bpwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=0
        )
        assert u.isclose(n_e_tot, n_e_tot_calc, atol=0 * u.Unit("cm-3"), rtol=1e-2)

    def test_from_normalised_energy_density(self):
        """test the intialisation of the powerlaw from the total particle 
        energy density"""
        u_e = 3e-4 * u.Unit("erg cm-3")
        bpwl = BrokenPowerLaw.from_normalised_energy_density(
            u_e=u_e,
            p1=p1_test,
            p2=p2_test,
            gamma_b=gamma_b_test,
            gamma_min=gamma_min_test,
            gamma_max=gamma_max_test,
        )
        # calculate u_e
        u_e_calc = mec2 * bpwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=1
        )
        assert u.isclose(u_e, u_e_calc, atol=0 * u.Unit("erg cm-3"), rtol=1e-2)

    def test_from_norm_at_gamma_1(self):
        """test the intialisation of the powerlaw from the normalisation at 
        gamma = 1"""
        norm = 1e-13 * u.Unit("cm-3")
        bpwl = BrokenPowerLaw.from_norm_at_gamma_1(
            norm=norm,
            p1=p1_test,
            p2=p2_test,
            gamma_b=gamma_b_test,
            gamma_min=1,
            gamma_max=gamma_max_test,
        )
        assert u.isclose(norm, bpwl(1), atol=0 * u.Unit("cm-3"), rtol=1e-2)


class TestLogParabola:
    """class grouping all tests related to the PowerLaw spectrum
    the analytical integral is non-trivial so we ignore the comparison with the 
    analytical integral"""

    def test_call(self):
        """assert that outside the bounding box the function returns 0"""
        gamma = np.logspace(0, 8)
        values = lp_test(gamma).value
        condition = (gamma_min_test <= gamma) * (gamma <= gamma_max_test)
        # check that outside the boundaries values are all 0
        assert not np.all(values[~condition])

    def test_from_normalised_density(self):
        """test the intialisation of the log parabola from the total particle 
        density"""
        n_e_tot = 1e-5 * u.Unit("cm-3")
        lp = LogParabola.from_normalised_density(
            n_e_tot=n_e_tot,
            p=p_test,
            q=q_test,
            gamma_0=gamma_0_test,
            gamma_min=gamma_min_test,
            gamma_max=gamma_max_test,
        )
        # calculate n_e_tot
        n_e_tot_calc = lp.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=0
        )
        assert u.isclose(n_e_tot, n_e_tot_calc, atol=0 * u.Unit("cm-3"), rtol=1e-2)

    def test_from_normalised_energy_density(self):
        """test the intialisation of the powerlaw from the total particle 
        energy density"""
        u_e = 3e-4 * u.Unit("erg cm-3")
        lp = LogParabola.from_normalised_energy_density(
            u_e=u_e,
            p=p_test,
            q=q_test,
            gamma_0=gamma_0_test,
            gamma_min=gamma_min_test,
            gamma_max=gamma_max_test,
        )
        # calculate u_e
        u_e_calc = mec2 * lp.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=1
        )
        assert u.isclose(u_e, u_e_calc, atol=0 * u.Unit("erg cm-3"), rtol=1e-2)

    def test_from_norm_at_gamma_1(self):
        """test the intialisation of the powerlaw from the normalisation at 
        gamma = 1"""
        norm = 1e-13 * u.Unit("cm-3")
        lp = LogParabola.from_norm_at_gamma_1(
            norm=norm,
            p=p_test,
            q=q_test,
            gamma_0=gamma_0_test,
            gamma_min=1,
            gamma_max=gamma_max_test,
        )
        assert u.isclose(norm, lp(1), atol=0 * u.Unit("cm-3"), rtol=1e-2)


class TestExpCutoffPowerLaw:
    """class grouping all tests related to the ExpCutoffPowerLaw spectrum"""

    def test_call(self):
        """assert that outside the bounding box the function returns 0"""
        gamma = np.logspace(0, 8)
        values = epwl_test(gamma).value
        condition = (gamma_min_test <= gamma) * (gamma <= gamma_max_test)
        # check that outside the boundaries values are all 0
        assert not np.all(values[~condition])

    def test_from_normalised_density(self):
        """test the intialisation of the power law from the total particle 
        density"""
        n_e_tot = 1e-5 * u.Unit("cm-3")
        epwl = ExpCutoffPowerLaw.from_normalised_density(
            n_e_tot=n_e_tot,
            p=p_test,
            gamma_c=gamma_c_test,
            gamma_min=gamma_min_test,
            gamma_max=gamma_max_test,
        )
        # calculate n_e_tot
        n_e_tot_calc = epwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=0
        )
        assert u.isclose(n_e_tot, n_e_tot_calc, atol=0 * u.Unit("cm-3"), rtol=1e-2)

    def test_from_normalised_energy_density(self):
        """test the intialisation of the power law from the total particle 
        energy density"""
        u_e = 3e-4 * u.Unit("erg cm-3")
        epwl = ExpCutoffPowerLaw.from_normalised_energy_density(
            u_e=u_e,
            p=p_test,
            gamma_c=gamma_c_test,
            gamma_min=gamma_min_test,
            gamma_max=gamma_max_test,
        )
        # calculate u_e
        u_e_calc = mec2 * epwl.integral(
            gamma_low=gamma_min_test, gamma_up=gamma_max_test, gamma_power=1
        )
        print(u_e)
        print(u_e_calc)
        assert u.isclose(u_e, u_e_calc, atol=0 * u.Unit("erg cm-3"), rtol=1e-2)

    def test_from_norm_at_gamma_1(self):
        """test the intialisation of the powerlaw from the normalisation at 
        gamma = 1"""
        norm = 1e-13 * u.Unit("cm-3")
        epwl = ExpCutoffPowerLaw.from_norm_at_gamma_1(
            norm=norm,
            p=p_test,
            gamma_c=gamma_c_test,
            gamma_min=1,
            gamma_max=gamma_max_test,
        )
        assert u.isclose(norm, epwl(1), atol=0 * u.Unit("cm-3"), rtol=1e-2)



import numpy as np
import astropy.units as u
from astropy.constants import e, c, m_e, sigma_T
from astropy.coordinates import Distance
import matplotlib.pyplot as plt
from .. import spectra
from ..utils.conversion import mec2, B_to_cgs
from ..utils.plot import load_mpl_rc

__all__ = ["Blob"]


def init_spectrum_norm_dict(norm, spectrum_dict, norm_type="integral", V_b=None):
    """initialize a spectrum from a normalisation and a spectrum dictionary of 
    the following type
    
    .. code-block:: python

        spectrum_dict = {
            "type": "PowerLaw", 
            "parameters": {
                "p": 2.8, 
                "gamma_min": 1e2, 
                "gamma_max": 1e7
            }
        }

    Parameters
    ----------
    norm : :class:`~astropy.units.Quantity`
        normalisation of the spectrum, might be in cm-3, erg cm-3 or erg,
        if erg are specified then V_b has to be provided
    spectrum_dict : dictionary
        dictionary with the spectrum type and parameters
    norm_type : ["integral", "differential", "gamma=1"]
        normalisation type
    V_b : :class:`~astropy.units.Quantity`
        volume of the emission region, to be provided only if the normalisation
        is in erg
    """
    model = getattr(spectra, spectrum_dict["type"])

    if norm.unit in (u.Unit("erg"), u.Unit("erg cm-3")) and norm_type != "integral":
        raise NameError(
            "Normalisations different than 'integral' available only for 'spectrum_norm' in cm-3"
        )

    # check the units of the normalisation
    # cm-3 is the only one allowing more than one normalisation type
    if norm.unit == u.Unit("cm-3"):
        if norm_type == "differential":
            final_model = model(norm, **spectrum_dict["parameters"])
        elif norm_type == "gamma=1":
            final_model = model.from_norm_at_gamma_1(
                norm, **spectrum_dict["parameters"]
            )
        elif norm_type == "integral":
            final_model = model.from_normalised_density(
                norm, **spectrum_dict["parameters"]
            )

    elif norm.unit == u.Unit("erg cm-3"):
        final_model = model.from_normalised_energy_density(
            norm, **spectrum_dict["parameters"]
        )

    elif norm.unit == u.Unit("erg"):
        if V_b is None:
            raise ValueError(
                "if a normalisation in erg is provided, the volume V_b of the emission region msut be specified"
            )
        else:
            final_model = model.from_total_energy(
                norm, V_b, **spectrum_dict["parameters"]
            )

    return final_model


class Blob:
    r"""Simple spherical emission region.

    **Note:** all these quantities are defined in the comoving frame so they are actually
    primed quantities, when referring the notation in [DermerMenon2009]_.

    Parameters
    ----------
    R_b : :class:`~astropy.units.Quantity`
        radius of the blob
    z : float
        redshift of the source
    delta_D : float
        Doppler factor of the relativistic outflow
    Gamma : float
        Lorentz factor of the relativistic outflow
    B : :class:`~astropy.units.Quantity`
        magnetic field in the blob (Gauss)
    xi : float
        acceleration coefficient :math:`\xi` for first-order Fermi acceleration
        :math:`(\mathrm{d}E/\mathrm{d}t \propto v \approx c)`
        used to compute limits on the maximum Lorentz factor via
        :math:`(\mathrm{d}E/\mathrm{d}t)_{\mathrm{acc}} = \xi c E / R_L`

    spectrum_norm : :class:`~astropy.units.Quantity`
        normalisation of the electron spectra, by default can be, following 
        the notation in [DermerMenon2009]_:

            - :math:`n_{e,\,tot}`: total electrons density, in :math:`\mathrm{cm}^{-3}`
            - :math:`u_e` : total electrons energy density, in :math:`\mathrm{erg}\,\mathrm{cm}^{-3}`
            - :math:`W_e` : total energy in electrons, in :math:`\mathrm{erg}`
        
        see `spectrum_norm_type` for more details on the normalisation

    spectrum_dict : dictionary
        dictionary containing type and spectral shape information, e.g.:

        .. code-block:: python

            spectrum_dict = {
                "type": "PowerLaw", 
                "parameters": {
                    "p": 2.8, 
                    "gamma_min": 1e2, 
                    "gamma_max": 1e7
                }
            }

    spectrum_norm_type : ["integral", "differential", "gamma=1"]
        only with a normalisation in "cm-3" one can select among three types: 

        * ``"integral"``: (default) the spectrum is set such that :math:`n_{e,\,tot}` equals the value provided by ``spectrum_norm``;  
        
        * ``"differential"``: the spectrum is set such that :math:`k_e` equals the value provided by ``spectrum_norm``;    
        
        * ``"gamma=1"``: the spectrum is set such that :math:`n_e(\gamma=1)` equals the value provided by ``spectrum_norm``.
        
    gamma_size : int
        size of the array of electrons Lorentz factors
    """

    def __init__(
        self,
        R_b=1e16 * u.cm,
        z=0.069,
        delta_D=10,
        Gamma=10,
        B=1 * u.G,
        spectrum_norm=1e48 * u.Unit("erg"),
        spectrum_dict={
            "type": "PowerLaw",
            "parameters": {"p": 2.8, "gamma_min": 1e2, "gamma_max": 1e7},
        },
        spectrum_norm_type="integral",
        xi=1.0,
        gamma_size=200,
    ):
        self.R_b = R_b.to("cm")
        self.z = z
        self.d_L = Distance(z=self.z).cgs
        self.V_b = 4 / 3 * np.pi * np.power(self.R_b, 3)
        self.delta_D = delta_D
        self.Gamma = Gamma
        self.Beta = np.sqrt(1 - 1 / np.power(self.Gamma, 2))
        # viewing angle
        self.mu_s = (1 - 1 / (self.Gamma * self.delta_D)) / self.Beta
        self.theta_s = (np.arccos(self.mu_s) * u.rad).to("deg")
        self.B = B
        # B decomposed in Gaussian-cgs units
        self.B_cgs = B_to_cgs(B)
        self.spectrum_norm = spectrum_norm
        self.spectrum_norm_type = spectrum_norm_type
        self.spectrum_dict = spectrum_dict
        self.xi = xi
        # default grid of Lorentz factors for integration in the external frame
        self.gamma_to_integrate = np.logspace(1, 9, gamma_size)
        # model for the electron density
        self.set_spectrum(
            self.spectrum_norm, self.spectrum_dict, self.spectrum_norm_type, gamma_size
        )

    def set_gamma_size(self, gamma_size):
        """change size of the array of electrons Lorentz factors"""
        self.gamma_size = gamma_size
        self.gamma = np.logspace(
            np.log10(self.gamma_min), np.log10(self.gamma_max), self.gamma_size
        )
        self.gamma_to_integrate = np.logspace(1, 9, self.gamma_size)

    def set_gamma(self, gamma_min, gamma_max, gamma_size):
        """set the array of Lorentz factors to be used for integration in the 
        frame comoving with the blob"""
        self.gamma_min = gamma_min
        self.gamma_max = gamma_max
        self.gamma_size = gamma_size
        # grid of Lorentz factor for the integration in the blob comoving frame
        self.gamma = np.logspace(
            np.log10(self.gamma_min), np.log10(self.gamma_max), self.gamma_size
        )

    def set_spectrum(
        self, spectrum_norm, spectrum_dict, spectrum_norm_type, gamma_size=200
    ):
        r"""set the spectrum :math:`n_e` for the electrons accelerated in the 
        blob, reset also the array of Lorentz factor given the `gamma_min` and
        `gamma_max` in the parameters dictionary"""
        self.set_gamma(
            spectrum_dict["parameters"]["gamma_min"],
            spectrum_dict["parameters"]["gamma_max"],
            gamma_size,
        )
        self.n_e = init_spectrum_norm_dict(
            spectrum_norm, spectrum_dict, spectrum_norm_type, self.V_b
        )

    def __str__(self):
        """printable summary of the blob"""
        return (
            "* spherical emission region\n"
            + f" - R_b (radius of the blob): {self.R_b.cgs:.2e}\n"
            + f" - V_b (volume of the blob): {self.V_b.cgs:.2e}\n"
            + f" - z (source redshift): {self.z:.2f}\n"
            + f" - d_L (source luminosity distance):{self.d_L.cgs:.2e}\n"
            + f" - delta_D (blob Doppler factor): {self.delta_D:.2e}\n"
            + f" - Gamma (blob Lorentz factor): {self.Gamma:.2e}\n"
            + f" - Beta (blob relativistic velocity): {self.Beta:.2e}\n"
            + f" - theta_s (jet viewing angle): {self.theta_s:.2e}\n"
            + f" - B (magnetic field tangled to the jet): {self.B:.2e}\n"
            + f" - xi (coefficient for 1st order Fermi acceleration) : {self.xi:.2e}\n"
            + str(self.n_e)
        )

    def set_delta_D(self, Gamma, theta_s):
        """set the viewing angle and the Lorentz factor of the outflow to
        obtain a specific Doppler factor

        Parameters
        ----------
        Gamma : float
            Lorentz factor of the relativistic outflow
        theta_s : :class:`~astropy.units.Quantity`
            viewing angle of the jet
        """
        mu_s = np.cos(theta_s.to("rad").value)
        Beta = np.sqrt(1 - 1 / np.power(Gamma, 2))
        delta_D = 1 / (Gamma * (1 - Beta * mu_s))

        self.theta_s = theta_s
        self.mu_s = mu_s
        self.Gamma = Gamma
        self.Beta = Beta
        self.delta_D = delta_D

    def N_e(self, gamma):
        r"""number of electrons as a function of the Lorentz factor, 
        :math:`N_e(\gamma') = V_b\,n_e(\gamma')`"""
        return self.V_b * self.n_e(gamma)

    @property
    def n_e_tot(self):
        r"""total electrons density

        .. math::
            n_{e,\,tot} = \int^{\gamma'_{\rm max}}_{\gamma'_{\rm min}} {\rm d}\gamma' n_e(\gamma')
        """
        return np.trapz(self.n_e(self.gamma), self.gamma)

    @property
    def N_e_tot(self):
        r"""total number of electrons

        .. math::
            N_{e,\,tot} = \int^{\gamma'_{\rm max}}_{\gamma'_{\rm min}} {\rm d}\gamma' N_e(\gamma')
        """
        return np.trapz(self.N_e(self.gamma), self.gamma)

    @property
    def u_e(self):
        r"""total energy density in non-thermal electrons

        .. math::
            u_{e} = m_e c^2\,\int^{\gamma'_{\rm max}}_{\gamma'_{\rm min}} {\rm d}\gamma' \gamma' n_e(\gamma')
        """
        return mec2 * np.trapz(self.gamma * self.n_e(self.gamma), self.gamma)

    @property
    def W_e(self):
        r"""total energy in non-thermal electrons

        .. math::
            W_{e} = m_e c^2\,\int^{\gamma'_{\rm max}}_{\gamma'_{\rm min}} {\rm d}\gamma' \gamma' N_e(\gamma')
        """
        return mec2 * np.trapz(self.gamma * self.N_e(self.gamma), self.gamma)

    @property
    def U_B(self):
        r"""energy density of magnetic field

        .. math::
            U_B = B^2 / (8 \pi)
        """
        U_B = np.power(self.B_cgs, 2) / (8 * np.pi)
        return U_B.to("erg cm-3")

    @property
    def k_eq(self):
        """equipartition parameter: ratio between totoal electron energy density  
        magnetic field energy density, Eq. 7.75 of [DermerMenon2009]_"""
        return (self.u_e / self.U_B).to_value("")

    @property
    def P_jet_e(self):
        r"""jet power in electrons

        .. math::
            P_{jet,\,e} = 2 \pi R_b^2 \beta \Gamma^2 c u_e
        """
        prefactor = (
            2 * np.pi * np.power(self.R_b, 2) * self.Beta * np.power(self.Gamma, 2) * c
        )
        return (prefactor * self.u_e).to("erg s-1")

    @property
    def P_jet_B(self):
        r"""jet power in magnetic field

        .. math::
            P_{jet,\,B} = 2 \pi R_b^2 beta \Gamma^2 c \frac{B^2}{8\pi}
        """
        prefactor = (
            2 * np.pi * np.power(self.R_b, 2) * self.Beta * np.power(self.Gamma, 2) * c
        )
        return (prefactor * self.U_B).to("erg s-1")

    @property
    def u_ph_synch(self):
        r"""energy density of the synchrotron photons energy losses are:

        .. math::        
            (\mathrm{d}E/\mathrm{d}t)_{\mathrm{synch}} = 4 / 3 \sigma_T c U_B \gamma^2 

        the radiation stays an average time of :math:`(3/4) (R_b/c)` (the factor of 3/4 cames from averaging over a sphere), 
        so an e- with Lorentz factor :math:`\gamma` produces:

        .. math::        
            0.75\,(\mathrm{d}E/\mathrm{d}t)_{\mathrm{synch}}\,(R_b/c)\,/\,V_b 

        of radiation. We need to integrate over the electron spectrum  (and multiply back by V_b)

        .. math::        
            0.75\,\int n_e(\gamma) (\mathrm{d}E/\mathrm{d}t)_{\mathrm{synch}}  R_b  \mathrm{d}\gamma
        
        so

        .. math::        
            u_{\mathrm{synch}} = \sigma_T  U_B  R_b  \int n_e(\gamma) \, \gamma^2 \mathrm{d}\gamma

        WARNING: this does not take into account SSA!
        """
        u_ph = (
            sigma_T.cgs
            * self.U_B
            * self.R_b
            * np.trapz(np.power(self.gamma, 2) * self.n_e(self.gamma), self.gamma)
        )
        return u_ph.to("erg cm-3")

    def plot_n_e(self, ax=None, gamma_power=0):
        """plot the  electron distribution
        
        Parameters 
        ----------
        ax : :class:`~matplotlib.axes.Axes`, optional
            Axis
        gamma_power : float
            power of gamma to raise the electron distribution
        """
        load_mpl_rc()
        ax = plt.gca() if ax is None else ax

        ax.loglog(self.gamma, np.power(self.gamma, gamma_power) * self.n_e(self.gamma))
        ax.set_xlabel(r"$\gamma$")
        if gamma_power == 0:
            ax.set_ylabel(r"$n_e(\gamma)\,/\,{\rm cm}^{-3}$")
        else:
            ax.set_ylabel(
                r"$\gamma^{"
                + str(gamma_power)
                + r"}$"
                + r"$\,n_e(\gamma)\,/\,{\rm cm}^{-3}$"
            )
        return ax



# math utilities for agnpy
import numpy as np
import warnings
import astropy.units as u

# default arrays to be used for integration
gamma_to_integrate = np.logspace(1, 9, 200)
nu_to_integrate = np.logspace(5, 30, 200) * u.Hz  # used for SSC
mu_to_integrate = np.linspace(-1, 1, 100)
phi_to_integrate = np.linspace(0, 2 * np.pi, 50)

# minimum relative distance to the absorber (to avoid infinite integrals)
min_rel_distance = 1.0e-4

# type of float to be used for math operation
numpy_type = np.float64
# smallest positive float
ftiny = np.finfo(numpy_type).tiny
# largest positive float
fmax = np.finfo(numpy_type).max


def axes_reshaper(*args):
    """reshape 1-dimensional arrays of different lengths in order for them to be
    broadcastable in multi-dimensional operations

    the rearrangement scheme for a list of n arrays is the following:
    `args[0]` is reshaped as `(args[0].size, 1, 1, ..., 1)` -> axis 0
    `args[1]` is reshaped as `(1, args[1].size, 1, ..., 1)` -> axis 1
    `args[2]` is reshaped as `(1, 1, args[2].size ..., 1)` -> axis 2
        .
        .
        .
    `args[n-1]` is reshaped as `(1, 1, 1, ..., args[n-1].size)` -> axis n-1
    
    Parameters
    ----------
    args: 1-dimensional `~numpy.ndarray`s to be reshaped
    """
    n = len(args)
    dim = (1,) * n
    reshaped_arrays = []
    for i, arg in enumerate(args):
        reshaped_dim = list(dim)  # the tuple is copied in the list
        reshaped_dim[i] = arg.size
        reshaped_array = np.reshape(arg, reshaped_dim)
        reshaped_arrays.append(reshaped_array)
    return reshaped_arrays


def log(x):
    """clipped log to avoid RuntimeWarning: divide by zero encountered in log"""
    values = np.clip(x, ftiny, fmax)
    return np.log(values)


def trapz_loglog(y, x, axis=0):
    """
    Integrate a function approximating its discrete intervals as power-laws
    (straight lines in log-log space), largely copied from naima

    Parameters
    ----------
    y : array_like
        array to integrate
    x : array_like, optional
        independent variable to integrate over
    axis : int, optional
        along which axis the integration has to be performed

    Returns
    -------
    trapz : float
        Definite integral as approximated by trapezoidal rule in loglog space.
    """
    try:
        y_unit = y.unit
        y = y.value
    except AttributeError:
        y_unit = 1.0
    try:
        x_unit = x.unit
        x = x.value
    except AttributeError:
        x_unit = 1.0

    slice_low = [slice(None)] * y.ndim
    slice_up = [slice(None)] * y.ndim
    # multi-dimensional equivalent of x_low = x[:-1]
    slice_low[axis] = slice(None, -1)
    # multi-dimensional equivalent of x_up = x[1:]
    slice_up[axis] = slice(1, None)

    # reshape x to be broadcastable with y
    if x.ndim == 1:
        shape = [1] * y.ndim
        shape[axis] = x.shape[0]
        x = x.reshape(shape)

    x_low = x[tuple(slice_low)]
    x_up = x[tuple(slice_up)]
    y_low = y[tuple(slice_low)]
    y_up = y[tuple(slice_up)]

    # slope in the given logarithmic bin
    m = (log(y_low) - log(y_up)) / (log(x_low) - log(x_up))

    vals = np.where(
        np.abs(m + 1) > 1e-10,
        y_low / (m + 1) * (x_up * np.power(x_up / x_low, m) - x_low),
        x_low * y_low * log(x_up / x_low),
    )

    tozero = (
        (y[tuple(slice_low)] == 0.0)
        + (y[tuple(slice_up)] == 0.0)
        + (x[tuple(slice_low)] == x[tuple(slice_up)])
    )
    vals[tozero] = 0.0

    return np.add.reduce(vals, axis) * x_unit * y_unit



# utils for testing
import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt

SED_X_LABEL = r"$\nu\,/\,{\rm Hz}$"
SED_Y_LABEL = r"$\nu F_{\nu}\,/\,({\rm erg}\,{\rm cm}^{-2}\,{\rm s}^{-1})$"
SED_DEVIATION_LABEL = r"$\nu F_{\nu, \rm agnpy}\,/\,\nu F_{\nu, \rm reference} - 1$"

TAU_X_LABEL = r"$\nu\,/\,{\rm Hz}$"
TAU_Y_LABEL = r"$\tau_{\gamma\gamma}$"
TAU_DEVIATION_LABEL = (
    r"$\tau_{\gamma\gamma, \rm agnpy}\,/\,\tau_{\gamma\gamma, \rm reference}$ - 1"
)


def extract_columns_sample_file(sample_file, x_unit, y_unit=None):
    """return two arrays of quantities from a sample file"""
    sample_table = np.loadtxt(sample_file, delimiter=",", comments="#")
    x = sample_table[:, 0] * u.Unit(x_unit)
    y = sample_table[:, 1] if y_unit is None else sample_table[:, 1] * u.Unit(y_unit)
    return x, y


def check_deviation(x, y_comp, y_ref, rtol, x_range=None):
    """check the deviation of two quantities within a given range of x
    when setting atol = 0 in np.allclose it will check that
    |a - b| <= rtol * |b|, that is |a / b - 1| <= rtol. 
    If we choose the agnpy values to be a and the reference (code ro figure from 
    the literature) to be b then |a / b - 1| will be positive when agnpy
    overestimates the reference (a > b) and negative when agnpy underestimates 
    the reference (a < b). 
    """
    if x_range is not None:
        condition = (x >= x_range[0]) * (x <= x_range[1])
        y_ref = y_ref[condition]
        y_comp = y_comp[condition]
    return np.allclose(y_comp, y_ref, atol=0, rtol=rtol)


def make_comparison_plot(
    nu,
    y_comp,
    y_ref,
    comp_label,
    ref_label,
    fig_title,
    fig_path,
    plot_type,
    y_range=None,
    comparison_range=None,
    x_scale="log",
    y_scale="log",
):
    """make a comparison plot, for SED or gamma-gamma absorption 
    between two different sources: a reference (literature or another code)
    and a comparison (usually the agnpy result)

    Parameters
    ----------
    nu: :class:`~astropy.units.Quantity`
        frequencies over which the comparison plot has to be plotted
    y_comp: :class:`~astropy.units.Quantity` or :class:`~numpy.ndarray`
        SED or gamma-gamma absorption to be compared (usually agnpy)
    y_ref: :class:`~astropy.units.Quantity` or :class:`~numpy.ndarray`
        reference SED or gamma-gamma absorption (from literature or another code)
    ref_label : `string`
        label of the reference model
    comp_label : `string`
        label of the comparison model
    fig_title : `string`
        upper title of the figure    
    fig_path : `string`
        path to save the figure
    plot_type : `{"sed", "tau", ...}`
        whether we are doing a comparison plot for a SED or an optical depth 
        if another string is specified it will be used for the y axis
    y_range : list of float
        lower and upper limit of the y axis limt
    comparison_range : list of float
        plot the range over which the residuals were checked 
    """
    if plot_type == "sed":
        # set the axes labels for an SED plot
        x_label = SED_X_LABEL
        y_label = SED_Y_LABEL
        deviation_label = SED_DEVIATION_LABEL
    elif plot_type == "tau":
        # set the axes labels for a tau plot
        x_label = TAU_X_LABEL
        y_label = TAU_Y_LABEL
        deviation_label = TAU_DEVIATION_LABEL
    else:
        # set a custom y label, keep the x-axis in frequency
        x_label = SED_X_LABEL
        y_label = plot_type
        deviation_label = f"({plot_type} agnpy / {plot_type} ref.) - 1"
    # make the plot
    fig, ax = plt.subplots(
        2,
        sharex=True,
        gridspec_kw={"height_ratios": [2, 1], "hspace": 0.05},
        figsize=(8, 6),
    )
    # plot the SEDs or TAUs in the upper panel
    # plot the reference sed with a continuous line and agnpy sed with a dashed one
    ax[0].loglog(nu, y_ref, marker=".", ls="-", color="k", lw=1.5, label=ref_label)
    ax[0].loglog(
        nu, y_comp, marker=".", ls="--", color="crimson", lw=1.5, label=comp_label
    )
    ax[0].set_ylabel(y_label)
    ax[0].set_title(fig_title)
    ax[0].legend(loc="best")
    if y_range is not None:
        ax[0].set_ylim(y_range)
    # plot the deviation in the bottom panel
    deviation = y_comp / y_ref - 1
    ax[1].axhline(0, ls="-", color="darkgray")
    ax[1].axhline(0.2, ls="--", color="darkgray")
    ax[1].axhline(-0.2, ls="--", color="darkgray")
    ax[1].axhline(0.3, ls=":", color="darkgray")
    ax[1].axhline(-0.3, ls=":", color="darkgray")
    ax[1].set_ylim([-0.5, 0.5])
    ax[1].semilogx(
        nu,
        deviation,
        marker=".",
        ls="--",
        color="crimson",
        lw=1.5,
        label=deviation_label,
    )
    ax[1].set_xlabel(x_label)
    ax[1].legend(loc="best")
    if comparison_range is not None:
        ax[1].axvline(comparison_range[0], ls="--", color="k")
        ax[1].axvline(comparison_range[1], ls="--", color="k")
    fig.savefig(f"{fig_path}")
    # avoid RuntimeWarning: More than 20 figures have been opened.
    plt.close(fig)
